# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-08-22 20:29
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('marketplace', '0022_section_is_standalone'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='offering',
            name='is_active',
        ),
        migrations.AddField(
            model_name='offering',
            name='state',
            field=django_fsm.FSMIntegerField(choices=[(1, 'Draft'), (2, 'Active'), (3, 'Paused'), (4, 'Archived')], default=1),
        ),
        migrations.AddField(
            model_name='offering',
            name='type',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
        migrations.AddField(
            model_name='plan',
            name='archived',
            field=models.BooleanField(default=False, help_text='Forbids creation of new resources.'),
        ),
        migrations.AddField(
            model_name='plan',
            name='article_code',
            field=models.CharField(blank=True, max_length=30),
        ),
        migrations.AddField(
            model_name='plan',
            name='content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='plan',
            name='object_id',
            field=models.PositiveIntegerField(null=True),
        ),
        migrations.AddField(
            model_name='plan',
            name='product_code',
            field=models.CharField(blank=True, max_length=30),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-08-28 11:31
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_vmware', '0021_virtualmachine_tools_state'),
    ]

    operations = [
        migrations.AddField(
            model_name='virtualmachine',
            name='tools_installed',
            field=models.BooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0014_instance_volumes'),
    ]

    operations = [
        migrations.AddField(
            model_name='instance',
            name='runtime_state',
            field=models.CharField(max_length=150, verbose_name='runtime state', blank=True),
        ),
    ]



import logging
import re
import sys

from django.db import IntegrityError
from django.utils import six, dateparse, timezone
from libcloud.common.types import LibcloudError
from libcloud.compute.drivers.ec2 import EC2NodeDriver, REGION_DETAILS, NAMESPACE, RESOURCE_EXTRA_ATTRIBUTES_MAP
from libcloud.compute.types import NodeState, StorageVolumeState
from libcloud.utils.xml import fixxpath

from waldur_core.core.models import SshPublicKey
from waldur_core.core.utils import hours_in_month
from waldur_core.structure import ServiceBackend, ServiceBackendError

from . import models

logger = logging.getLogger(__name__)


RESOURCE_EXTRA_ATTRIBUTES_MAP['volume']['volume_type'] = {
    'xpath': 'volumeType',
    'transform_func': str
}


class ExtendedEC2NodeDriver(EC2NodeDriver):
    def get_node(self, node_id):
        """
        Get a node based on an node_id

        :param node_id: Node identifier
        :type node_id: ``str``

        :return: A Node object
        :rtype: :class:`Node`
        """

        return self.list_nodes(ex_node_ids=[node_id])[0]

    def list_volumes(self, node_id=None, ex_volume_ids=None, ex_filters=None):
        """
        List all volumes

        ex_volume_ids parameter is used to filter the list of
        volumes that should be returned. Only the volumes
        with the corresponding volume ids will be returned.

        :param      node_id: Node identifier
        :type       node_id: ``str``

        :param      ex_volume_ids: List of ``volume.id``
        :type       ex_volume_ids: ``list`` of ``str``

        :param      ex_filters: The filters so that the response includes
                                information for only certain volumes.
        :type       ex_filters: ``dict``

        :rtype: ``list`` of :class:`Node`
        """

        params = {
            'Action': 'DescribeVolumes',
        }
        if node_id:
            filters = {'attachment.instance-id': node_id}
            params.update(self._build_filters(filters))

        if ex_volume_ids:
            params.update(self._pathlist('VolumeId', ex_volume_ids))

        if ex_filters:
            params.update(self._build_filters(ex_filters))

        response = self.connection.request(self.path, params=params).object
        volumes = [self._to_volume(el) for el in response.findall(
            fixxpath(xpath='volumeSet/item', namespace=NAMESPACE))
        ]
        return volumes

    def get_volume(self, volume_id):
        """
        Get a volume based on an volume_id

        :param volume_id: Volume identifier
        :type volume_id: ``str``

        :return: A Volume object
        :rtype: :class:`Volume`
        """
        return self.list_volumes(ex_volume_ids=[volume_id])[0]

    # Location is required argument
    # http://docs.aws.amazon.com/AWSEC2/latest/APIReference/API_CreateVolume.html
    def create_volume(self, size, name, location, snapshot=None,
                      ex_volume_type='standard', ex_iops=None):
        """
        Create a new volume.

        :param size: Size of volume in gigabytes (required)
        :type size: ``int``

        :param name: Name of the volume to be created
        :type name: ``str``

        :param snapshot:  Snapshot from which to create the new
                               volume.  (optional)
        :type snapshot:  :class:`.VolumeSnapshot`

        :param location: Datacenter in which to create a volume in (required).
        :type location: :class:`.ExEC2AvailabilityZone`

        :param ex_volume_type: Type of volume to create.
        :type ex_volume_type: ``str``

        :param iops: The number of I/O operations per second (IOPS)
                     that the volume supports. Only used if ex_volume_type
                     is io1.
        :type iops: ``int``

        :return: The newly created volume.
        :rtype: :class:`StorageVolume`
        """
        valid_volume_types = ['standard', 'io1', 'gp2']

        params = {
            'Action': 'CreateVolume',
            'Size': str(size)}

        if ex_volume_type and ex_volume_type not in valid_volume_types:
            raise ValueError('Invalid volume type specified: %s' %
                             (ex_volume_type))

        if snapshot:
            params['SnapshotId'] = snapshot.id

        params['AvailabilityZone'] = location.name

        if ex_volume_type:
            params['VolumeType'] = ex_volume_type

        if ex_volume_type == 'io1' and ex_iops:
            params['Iops'] = ex_iops

        volume = self._to_volume(
            self.connection.request(self.path, params=params).object,
            name=name)

        if self.ex_create_tags(volume, {'Name': name}):
            volume.extra['tags']['Name'] = name

        return volume

    def ex_change_node_size(self, node, size_id):
        """
        Change the node size.
        Note: Node must be turned of before changing the size.

        :param      node: Node instance
        :type       node: :class:`Node`

        :param      size_id: New size ID
        :type       size_id: string

        :return: True on success, False otherwise.
        :rtype: ``bool``
        """
        attributes = {'InstanceType.Value': size_id}
        return self.ex_modify_instance_attribute(node, attributes)


class AWSBackendError(ServiceBackendError):
    pass


def reraise(exc):
    """
    Reraise AWSBackendError while maintaining traceback.
    """
    six.reraise(AWSBackendError, exc, sys.exc_info()[2])


class AWSBackend(ServiceBackend):
    """ Waldur interface to AWS EC2 API.
        https://libcloud.apache.org/
    """
    State = NodeState
    Regions = (('us-east-1', 'US East (N. Virginia)'),
               ('us-west-2', 'US West (Oregon)'),
               ('us-west-1', 'US West (N. California)'),
               ('eu-west-1', 'EU (Ireland)'),
               ('eu-central-1', 'EU (Frankfurt)'),
               ('ap-south-1', 'Asia Pacific (Mumbai)'),
               ('ap-southeast-1', 'Asia Pacific (Singapore)'),
               ('ap-southeast-2', 'Asia Pacific (Sydney)'),
               ('ap-northeast-1', 'Asia Pacific (Tokyo)'),
               ('ap-northeast-2', 'Asia Pacific (Seoul)'),
               ('sa-east-1', 'South America (Sao Paulo)'))

    def __init__(self, settings):
        super(AWSBackend, self).__init__(settings)
        self.settings = settings

    def _get_api(self, region='us-east-1'):
        return ExtendedEC2NodeDriver(
            self.settings.username, self.settings.token, region=region)

    def ping(self, raise_exception=False):
        try:
            self._get_api().list_key_pairs()
        except Exception as e:
            if raise_exception:
                reraise(e)
            return False
        else:
            return True

    def ping_resource(self, instance):
        try:
            manager = self.get_manager(instance)
            manager.get_node(instance.backend_id)
        except LibcloudError:
            return False
        else:
            return True

    def pull_service_properties(self):
        self.pull_regions()
        self.pull_sizes()
        self.update_images()

    def pull_regions(self):
        nc_regions = set(models.Region.objects.values_list('backend_id', flat=True))
        for backend_id, name in self.Regions:
            if backend_id in nc_regions:
                continue
            try:
                models.Region.objects.create(backend_id=backend_id, name=name)
            except IntegrityError:
                message = 'Could not create AWS region with name %s due to concurrent update'
                logger.warning(message, name)

    def pull_sizes(self):
        regions = models.Region.objects.values_list('backend_id', flat=True)
        for region in regions:
            manager = self._get_api(region)

            # XXX: Obviously each region has a different price,
            #      find a better form of models relation
            for backend_size in manager.list_sizes():
                size, _ = models.Size.objects.update_or_create(
                    backend_id=backend_size.id,
                    defaults={
                        'name': backend_size.name,
                        'cores': backend_size.extra.get('cpu', 1),
                        'ram': self.gb2mb(backend_size.ram),
                        'disk': self.gb2mb(backend_size.disk),
                        'price': backend_size.price,
                    })

                current_regions = set(size.regions.all())
                backend_regions = set(models.Region.objects.filter(backend_id__in=[
                    r for r, v in REGION_DETAILS.items() if backend_size.id in v['instance_types']]))

                size.regions.add(*(backend_regions - current_regions))
                size.regions.remove(*(current_regions - backend_regions))

    def pull_images(self):
        cur_images = {i.backend_id: i for i in models.Image.objects.all()}

        for region, backend_image in self.get_all_images():
            cur_images.pop(backend_image.id, None)
            try:
                models.Image.objects.update_or_create(
                    backend_id=backend_image.id,
                    defaults={
                        'name': backend_image.name,
                        'region': region
                    })
            except IntegrityError:
                logger.warning(
                    'Could not create AWS image with id %s due to concurrent update',
                    backend_image.id)

        # Remove stale images using one SQL query
        models.Image.objects.filter(backend_id__in=cur_images.keys()).delete()

    def create_volume(self, volume):
        try:
            manager = self._get_api(volume.region.backend_id)
            zones = manager.ex_list_availability_zones()
            # Availability zone is required by AWS
            new_volume = manager.create_volume(
                location=zones[0],
                size=volume.size,
                name=volume.name,
                ex_volume_type=volume.volume_type
            )
            volume.backend_id = new_volume.id
            volume.save(update_fields=['backend_id'])
        except Exception as e:
            logger.exception('Unable to create volume with id %s', volume.id)
            reraise(e)

    def delete_volume(self, volume):
        try:
            manager = self._get_api(volume.region.backend_id)
            manager.destroy_volume(self.get_volume(volume))
        except Exception as e:
            logger.exception('Unable to delete volume with id %s', volume.id)
            reraise(e)

    def attach_volume(self, volume):
        """
        Attach volume to the instance
        """
        try:
            manager = self._get_api(volume.region.backend_id)
            backend_node = manager.get_node(volume.instance.backend_id)
            backend_volume = manager.get_volume(volume.backend_id)
            manager.attach_volume(backend_node, backend_volume, volume.device)
        except Exception as e:
            logger.exception('Unable to attach volume with id %s to instance with id %s',
                             volume.id, volume.instance.id)
            reraise(e)

    def detach_volume(self, volume):
        """
        Detach volume from the instance
        """
        try:
            manager = self._get_api(volume.region.backend_id)
            backend_volume = manager.get_volume(volume.backend_id)
            manager.detach_volume(backend_volume)
        except Exception as e:
            logger.exception('Unable to detach volume with id %s', volume.id)
            reraise(e)
        else:
            volume.instance = None
            volume.device = ''
            volume.save(update_fields=['instance', 'device'])

    def get_all_images(self):
        """
        Fetch images from all regions
        """
        # TODO: change into a more flexible filtering
        options = self.settings.options or {}
        regex = None
        if 'images_regex' in options:
            try:
                regex = re.compile(options['images_regex'])
            except re.error:
                logger.warning(
                    'Invalid images regexp supplied for service settings %s: %s',
                    self.settings.uuid, options['images_regex'])

        for region in models.Region.objects.all():
            manager = self._get_api(region.backend_id)
            # opinionated filter for populating image list
            for image in manager.list_images(ex_owner='aws-marketplace',
                                             ex_filters={'virtualization-type': 'hvm', 'image-type': 'machine'}):
                # Skip images without name
                if image.name:
                    if regex and not regex.match(image.name):
                        continue
                    yield region, image

    def update_images(self):
        def get_images(manager, owner):
            return {i.id: i.extra['description']
                    for i in manager.list_images(
                        ex_owner=owner,
                        ex_filters={'virtualization-type': 'hvm', 'image-type': 'machine'})
                    }

        for region in models.Region.objects.all():
            images = region.image_set.all()
            if images.count():
                manager = self._get_api(region.backend_id)
                backend_images = get_images(manager, 'amazon')
                backend_images.update(get_images(manager, 'aws-marketplace'))
                for image in images:
                    try:
                        name = backend_images[image.backend_id]
                        # Backend can return image with ID, but without name.
                        if name is None:
                            image.delete()
                            continue
                    except KeyError:
                        image.delete()
                    else:
                        image.name = name
                        image.save(update_fields=['name'])

    def get_all_nodes(self):
        """
        Fetch nodes from all regions
        """
        try:
            for region in models.Region.objects.all():
                manager = self._get_api(region.backend_id)
                for node in manager.list_nodes():
                    yield region, node
        except LibcloudError as e:
            reraise(e)

    def create_instance(self, instance, backend_image_id=None, backend_size_id=None, ssh_key_uuid=None):
        manager = self.get_manager(instance)

        params = dict(name=instance.name,
                      image=self.get_image(backend_image_id, manager),
                      size=self.get_size(backend_size_id, manager),
                      ex_custom_data=instance.user_data,
                      # Set volume termination on instance delete
                      ex_blockdevicemappings=[{
                          'DeviceName': '/dev/sda1',
                          'Ebs': {'DeleteOnTermination': 'true'}
                      }])

        if ssh_key_uuid:
            ssh_key = SshPublicKey.objects.get(uuid=ssh_key_uuid)
            try:
                backend_ssh_key = self.get_or_create_ssh_key(ssh_key, manager)
            except LibcloudError as e:
                logger.exception('Unable to provision SSH key %s', ssh_key_uuid)
                reraise(e)

            params['ex_keyname'] = backend_ssh_key['keyName']

        try:
            backend_instance = manager.create_node(**params)
        except LibcloudError as e:
            logger.exception('Failed to provision virtual machine %s', instance.name)
            reraise(e)

        if ssh_key_uuid:
            instance.key_name = ssh_key.name
            instance.key_fingerprint = ssh_key.fingerprint

        instance.backend_id = backend_instance.id
        instance.save(update_fields=['backend_id'])
        return instance

    def pull_instance_volume(self, volume):
        instance = volume.instance
        try:
            manager = self.get_manager(instance)
            backend_volume = manager.list_volumes(instance.backend_id)[0]
        except Exception as e:
            logger.exception('Failed to get volume for Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))

        volume.name = ('volume-%s' % instance.name)[:150]
        volume.backend_id = backend_volume.id
        volume.device = backend_volume.extra['device']
        volume.size = backend_volume.size
        volume.volume_type = backend_volume.extra['type']
        volume.save(update_fields=['name', 'backend_id', 'device', 'size', 'volume_type'])

    def reboot_instance(self, instance):
        try:
            manager = self.get_manager(instance)
            manager.reboot_node(manager.get_node(instance.backend_id))
        except Exception as e:
            logger.exception('Unable to reboot Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))
        else:
            instance.start_time = timezone.now()
            instance.save(update_fields=['start_time'])

    def stop_instance(self, instance):
        try:
            manager = self.get_manager(instance)
            manager.ex_stop_node(manager.get_node(instance.backend_id))
        except Exception as e:
            logger.exception('Unable to stop Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))
        else:
            instance.start_time = None
            instance.save(update_fields=['start_time'])

    def start_instance(self, instance):
        try:
            manager = self.get_manager(instance)
            manager.ex_start_node(manager.get_node(instance.backend_id))
        except Exception as e:
            logger.exception('Unable to start Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))
        else:
            instance.start_time = timezone.now()
            instance.save(update_fields=['start_time'])

    def destroy_instance(self, instance):
        try:
            manager = self.get_manager(instance)
            manager.destroy_node(manager.get_node(instance.backend_id))
        except Exception as e:
            logger.exception('Unable to destroy Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))
        else:
            instance.decrease_backend_quotas_usage()

    def resize_instance(self, instance, size_id):
        try:
            manager = self.get_manager(instance)
            manager.ex_change_node_size(manager.get_node(instance.backend_id), size_id)
        except Exception as e:
            logger.exception('Unable to resize Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))

    def pull_instance_runtime_state(self, instance):
        try:
            manager = self.get_manager(instance)
            backend_vm = manager.get_node(instance.backend_id)
        except Exception as e:
            logger.exception('Unable to pull state for Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))

        if backend_vm.state != instance.runtime_state:
            instance.runtime_state = backend_vm.state
            instance.save(update_fields=['runtime_state'])

    def pull_instance_public_ips(self, instance):
        try:
            manager = self.get_manager(instance)
            backend_vm = manager.get_node(instance.backend_id)
        except Exception as e:
            logger.exception('Unable to pull public IPs for Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))

        if backend_vm.public_ips != instance.public_ips:
            instance.public_ips = backend_vm.public_ips
            instance.save(update_fields=['public_ips'])

    def is_instance_terminated(self, instance):
        try:
            manager = self.get_manager(instance)
            backend_vm = manager.get_node(instance.backend_id)
        except Exception as e:
            logger.exception('Unable to check state for Amazon virtual machine %s', instance.uuid.hex)
            six.reraise(AWSBackendError, six.text_type(e))

        return backend_vm.state == self.State.TERMINATED

    def get_monthly_cost_estimate(self, instance):
        manager = self.get_manager(instance)
        try:
            backend_instance = manager.get_node(instance.backend_id)
        except Exception as e:
            reraise(e)

        size = self.get_size(backend_instance.extra['instance_type'], manager)

        # calculate a price for current month based on hourly rate
        return size.price * hours_in_month()

    def to_instance(self, instance, region):
        from waldur_core.structure import SupportedServices

        manager = self._get_api(region.backend_id)
        # TODO: Connect volume with instance
        try:
            volumes = {v.id: v.size for v in manager.list_volumes(instance.id)}
        except Exception as e:
            reraise(e)

        for device in instance.extra['block_device_mapping']:
            vid = device['ebs']['volume_id']
            if vid in volumes:
                device['ebs']['volume_size'] = volumes[vid]

        # libcloud is a funny buggy thing, put all required info here
        instance_type = self.get_size(instance.extra['instance_type'], manager)

        return {
            'id': instance.id,
            'name': instance.name or instance.uuid,
            'cores': instance_type.extra.get('cpu', 1),
            'ram': instance_type.ram,
            'disk': self.gb2mb(sum(volumes.values())),
            'created': dateparse.parse_datetime(instance.extra['launch_time']),
            'region': region.uuid.hex,
            'state': models.Instance.States.OK,
            'public_ips': instance.public_ips,
            'flavor_name': instance.extra.get('instance_type'),
            'type': SupportedServices.get_name_for_model(models.Instance),
            'runtime_state': instance.state
        }

    def get_manager(self, instance):
        return self._get_api(instance.region.backend_id)

    def get_size(self, size_id, manager):
        try:
            return next(s for s in manager.list_sizes() if s.id == size_id)
        except (StopIteration, LibcloudError) as e:
            logger.exception("Size %s doesn't exist", size_id)
            reraise(e)

    def get_image(self, image_id, manager):
        try:
            return manager.get_image(image_id)
        except (StopIteration, LibcloudError) as e:
            logger.exception("Image %s doesn't exist", image_id)
            reraise(e)

    def get_or_create_ssh_key(self, ssh_key, manager):
        try:
            return manager.ex_describe_keypair(ssh_key.name)
        except LibcloudError:
            return manager.ex_import_keypair_from_string(ssh_key.name, ssh_key.public_key)

    def get_resources_for_import(self, resource_type=None):
        from waldur_core.structure import SupportedServices

        resources = []

        if resource_type is None or resource_type == SupportedServices.get_name_for_model(models.Instance):
            resources.extend(self.get_instances_for_import())

        if resource_type is None or resource_type == SupportedServices.get_name_for_model(models.Volume):
            resources.extend(self.get_volumes_for_import())
        return resources

    def get_instances_for_import(self):
        cur_instances = models.Instance.objects.all().values_list('backend_id', flat=True)

        return [
            self.to_instance(instance, region)
            for region, instance in self.get_all_nodes()
            if instance.id not in cur_instances and
            instance.state != NodeState.TERMINATED
        ]

    def get_volumes_for_import(self):
        cur_volumes = models.Volume.objects.all().values_list('backend_id', flat=True)
        return [
            self.to_volume(volume)
            for region, volume in self.get_all_volumes()
            if volume.id not in cur_volumes and
            volume.state != StorageVolumeState.DELETED
        ]

    def find_instance(self, instance_id):
        for region in models.Region.objects.all():
            manager = self._get_api(region.backend_id)
            try:
                instance = manager.get_node(instance_id)
            except LibcloudError:
                # Instance not found
                pass
            else:
                return region, self.to_instance(instance, region)
        raise AWSBackendError("Instance with id %s is not found", instance_id)

    def find_volume(self, volume_id):
        for region in models.Region.objects.all():
            manager = self._get_api(region.backend_id)
            try:
                volume = manager.get_volume(volume_id)
            except LibcloudError:
                # Volume not found
                pass
            else:
                return region, self.to_volume(volume)
        raise AWSBackendError("Volume with id %s is not found", volume_id)

    def get_managed_resources(self):
        backend_instance = self.get_managed_instances()
        backend_volumes = self.get_managed_volumes()
        return list(backend_instance) + list(backend_volumes)

    def get_managed_instances(self):
        try:
            ids = [instance.id for region, instance in self.get_all_nodes()]
            return models.Instance.objects.filter(backend_id__in=ids)
        except LibcloudError:
            return []

    def get_managed_volumes(self):
        try:
            ids = [volume.id for region, volume in self.get_all_volumes()]
            return models.Volume.objects.filter(backend_id__in=ids)
        except LibcloudError:
            return []

    def get_all_volumes(self):
        try:
            for region in models.Region.objects.all():
                manager = self._get_api(region.backend_id)
                for node in manager.list_volumes():
                    yield region, node
        except Exception as e:
            logger.exception('Unable to list EC2 volumes')
            reraise(e)

    def to_volume(self, volume):
        from waldur_core.structure import SupportedServices

        return {
            'id': volume.id,
            'name': volume.name,
            'size': volume.size,
            'created': volume.extra['create_time'],
            'state': self._get_volume_state(volume.state),
            'runtime_state': volume.state,
            'type': SupportedServices.get_name_for_model(models.Volume),
            'device': volume.extra['device'],
            'instance_id': volume.extra['instance_id'],
            'volume_type': volume.extra['volume_type']
        }

    def _get_volume_state(self, state):
        aws_to_waldur = {
            StorageVolumeState.AVAILABLE: models.Volume.States.OK,
            StorageVolumeState.INUSE: models.Volume.States.OK,
            StorageVolumeState.CREATING: models.Volume.States.CREATING,
            StorageVolumeState.DELETING: models.Volume.States.DELETING,
            StorageVolumeState.ATTACHING: models.Volume.States.UPDATING
        }

        return aws_to_waldur.get(state, models.Volume.States.ERRED)

    def get_volume(self, volume):
        try:
            manager = self._get_api(volume.region.backend_id)
            return manager.get_volume(volume.backend_id)
        except LibcloudError as e:
            reraise(e)

    def pull_volume_runtime_state(self, volume):
        backend_volume = self.get_volume(volume)
        if backend_volume.state != volume.runtime_state:
            volume.runtime_state = backend_volume.state
            volume.save(update_fields=['runtime_state'])



import django_filters
from waldur_core.core import filters as core_filters

from . import models


class InvoiceFilter(django_filters.FilterSet):
    customer = core_filters.URLFilter(view_name='customer-detail', name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(name='customer__uuid')
    state = django_filters.MultipleChoiceFilter(choices=models.Invoice.States.CHOICES)
    o = django_filters.OrderingFilter(fields=(('year', 'month'),))

    class Meta(object):
        model = models.Invoice
        fields = ('year', 'month')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-11-02 08:11
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('support', '0030_issue_template'),
    ]

    operations = [
        migrations.CreateModel(
            name='SupportCustomer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('backend_id', models.CharField(max_length=255, unique=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-23 12:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0028_ignoredissuestatus_templatestatusnotification'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='resolution_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-22 12:41
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0051_resource_backend_metadata'),
    ]

    operations = [
        migrations.CreateModel(
            name='ResourcePlanPeriod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('start', models.DateTimeField(blank=True, null=True, verbose_name='start')),
                ('end', models.DateTimeField(blank=True, null=True, verbose_name='end')),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace.Plan')),
                ('resource', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='marketplace.Resource')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-20 15:32
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0015_rename_order_item'),
    ]

    operations = [
        migrations.CreateModel(
            name='AttributeOption',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=255)),
                ('title', models.CharField(max_length=255)),
            ],
        ),
        migrations.RemoveField(
            model_name='attribute',
            name='available_values',
        ),
        migrations.AddField(
            model_name='attributeoption',
            name='attribute',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='options', to='marketplace.Attribute'),
        ),
        migrations.AlterUniqueTogether(
            name='attributeoption',
            unique_together=set([('attribute', 'key')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-11-20 13:59
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_fsm


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0046_migrate_resource'),
    ]

    operations = [
        migrations.AlterField(
            model_name='componentquota',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='quotas',
                                    to='marketplace.Resource'),
        ),
        migrations.AlterField(
            model_name='componentusage',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='usages',
                                    to='marketplace.Resource'),
        ),
        migrations.AlterField(
            model_name='orderitem',
            name='state',
            field=django_fsm.FSMIntegerField(choices=[(1, 'pending'), (2, 'executing'), (3, 'done'), (4, 'erred'),
                                                      (5, 'terminated'), (6, 'terminating')], default=1),
        ),
        migrations.AlterField(
            model_name='resource',
            name='offering',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+',
                                    to='marketplace.Offering'),
        ),
        migrations.AlterUniqueTogether(
            name='componentquota',
            unique_together=set([('resource', 'component')]),
        ),
        migrations.AlterUniqueTogether(
            name='componentusage',
            unique_together=set([('resource', 'component', 'date')]),
        ),
        migrations.RemoveField(
            model_name='orderitem',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='orderitem',
            name='object_id',
        ),
        migrations.RemoveField(
            model_name='componentusage',
            name='order_item',
        ),
        migrations.RemoveField(
            model_name='componentquota',
            name='order_item',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-08-27 09:32
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0092_improve_internal_name_validation'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringcomponent',
            name='max_value',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='offeringcomponent',
            name='min_value',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-10-03 12:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_rancher', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cluster',
            name='backend_id',
            field=models.CharField(blank=True, default='', max_length=255),
            preserve_default=False,
        ),
    ]



import decimal

from django.conf import settings

from waldur_mastermind.invoices.registrators import BaseRegistrator
from waldur_mastermind.invoices import models as invoices_models
from waldur_mastermind.marketplace_openstack import PACKAGE_TYPE, CORES_TYPE, RAM_TYPE, STORAGE_TYPE

from waldur_mastermind.marketplace import models
from waldur_mastermind.common.utils import mb_to_gb


class MarketplaceItemRegistrator(BaseRegistrator):
    def get_customer(self, source):
        return source.project.customer

    def get_sources(self, customer):
        if not settings.WALDUR_MARKETPLACE_OPENSTACK['BILLING_ENABLED']:
            return models.Resource.objects.none()

        return models.Resource.objects.filter(
            project__customer=customer,
            offering__type=PACKAGE_TYPE
        ).exclude(state__in=[
            models.Resource.States.CREATING,
            models.Resource.States.TERMINATED
        ])

    def _create_item(self, source, invoice, start, end):
        from waldur_mastermind.marketplace import plugins

        details = self.get_details(source)
        builtin_components = plugins.manager.get_components(source.offering.type)
        component_factors = {c.type: c.factor for c in builtin_components}
        unit_price = sum(
            decimal.Decimal(source.limits.get(component.component.type, 0)) * component.price /
            decimal.Decimal(component_factors.get(component.component.type, 1))
            for component in source.plan.components.all()
        )

        start = invoices_models.adjust_invoice_items(
            invoice, source, start, unit_price, source.plan.unit)

        item = invoices_models.InvoiceItem.objects.create(
            scope=source,
            project=source.project,
            unit_price=unit_price,
            unit=source.plan.unit,
            product_code=source.plan.product_code,
            article_code=source.plan.article_code,
            invoice=invoice,
            start=start,
            end=end,
            details=details,
        )
        self.init_details(item)

    def get_name(self, source):
        return '{resource} ({offering} / VPC {cores} CPU - {ram} GB RAM - {disk} GB storage)'.format(
            resource=source.name,
            offering=source.offering.name,
            cores=source.limits.get(CORES_TYPE),
            ram=int(mb_to_gb(source.limits.get(RAM_TYPE, 0))),
            disk=int(mb_to_gb(source.limits.get(STORAGE_TYPE, 0))),
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0002_invoice_additional_details'),
    ]

    operations = [
        migrations.AddField(
            model_name='paymentdetails',
            name='default_tax_percent',
            field=models.DecimalField(default=0, max_digits=4, decimal_places=2, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)]),
        ),
    ]



"""Exceptions module for VMware vCenter Automation API client."""


class VMwareError(Exception):
    """Base exception class for VMware client."""

    message = 'An unknown exception occurred.'

    def __init__(self, message=None, code=None):
        """Initialize exception class with optional message and code."""
        if message:
            self.message = message
        if code:
            self.code = code

    def __str__(self):
        """Serialize exception to string using it's message."""
        return self.message


class BadRequest(VMwareError):
    """General purpose exception class."""


class Unauthorized(BadRequest):
    """Raised when invalid credentials are provided."""

    message = 'Unauthorized: bad credentials.'



import django_filters
from rest_framework.filters import BaseFilterBackend

from waldur_core.core import filters as core_filters
from waldur_core.core.utils import get_ordering
from waldur_core.structure import models as structure_models

from .models import ResourceSlaStateTransition
from .utils import get_period


class ResourceScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return structure_models.ResourceMixin.get_all_models()

    def get_field_name(self):
        return 'scope'


class ResourceStateFilter(django_filters.FilterSet):
    class Meta:
        model = ResourceSlaStateTransition
        fields = ('timestamp', 'period', 'state')


class SlaFilter(BaseFilterBackend):
    """
    SLA filter
    ^^^^^^^^^^

    Allows to filter or sort resources by actual_sla
    Default period is current year and month.

    Example query parameters for filtering list of OpenStack instances:

    .. code-block:: http

        /api/openstack-instances/?actual_sla=90&period=2016-02

    Example query parameters for sorting list of OpenStack instances:

    .. code-block:: http

        /api/openstack-instances/?o=actual_sla&period=2016-02
    """

    def filter_queryset(self, request, queryset, view):
        period = get_period(request)

        if 'actual_sla' in request.query_params:
            value = request.query_params.get('actual_sla')
            return queryset.filter(sla_items__value=value, sla_items__period=period)

        elif get_ordering(request) == 'actual_sla':
            return queryset.filter(sla_items__period=period).order_by('sla_items__value')

        else:
            return queryset


class MonitoringItemFilter(BaseFilterBackend):
    """
    Monitoring filter
    ^^^^^^^^^^^^^^^^^

    Filter and order resources by monitoring item.
    For example, given query dictionary

    .. code-block:: http

        {
            'monitoring__installation_state': True
        }

    it produces following query

    .. code-block:: http

        {
            'monitoring_item__name': 'installation_state',
            'monitoring_item__value': True
        }

    Example query parameters for sorting list of OpenStack instances:

    .. code-block:: http

        /api/openstack-instances/?o=monitoring__installation_state
    """

    def filter_queryset(self, request, queryset, view):
        for key in request.query_params.keys():
            item_name = self._get_item_name(key)
            if item_name:
                value = request.query_params.get(key)
                queryset = queryset.filter(monitoring_items__name=item_name,
                                           monitoring_items__value=value)

        order_by = get_ordering(request)
        item_name = self._get_item_name(order_by)
        if item_name:
            queryset = queryset.filter(monitoring_items__name=item_name)\
                               .order_by('monitoring_items__value')

        return queryset

    def _get_item_name(self, key):
        if key and key.startswith('monitoring__'):
            _, item_name = key.split('__', 1)
            return item_name



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-06-19 18:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0088_fill_billing_period'),
    ]

    operations = [
        migrations.AlterField(
            model_name='componentusage',
            name='billing_period',
            field=models.DateField(),
        ),
    ]



from . import views


def register_in(router):
    router.register(r'aws', views.AmazonServiceViewSet, base_name='aws')
    router.register(r'aws-regions', views.RegionViewSet, base_name='aws-region')
    router.register(r'aws-images', views.ImageViewSet, base_name='aws-image')
    router.register(r'aws-sizes', views.SizeViewSet, base_name='aws-size')
    router.register(r'aws-instances', views.InstanceViewSet, base_name='aws-instance')
    router.register(r'aws-volumes', views.VolumeViewSet, base_name='aws-volume')
    router.register(r'aws-service-project-link',
                    views.AmazonServiceProjectLinkViewSet, base_name='aws-spl')



from __future__ import unicode_literals

from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin
from django.views.generic import TemplateView

from waldur_core.core import WaldurExtension
from waldur_core.core import views as core_views
from waldur_core.core.routers import SortedDefaultRouter as DefaultRouter
from waldur_core.core.schemas import WaldurSchemaView
from waldur_core.cost_tracking import urls as cost_tracking_urls, CostTrackingRegister
from waldur_core.logging import urls as logging_urls
from waldur_core.monitoring import urls as monitoring_urls
from waldur_core.quotas import urls as quotas_urls
from waldur_core.structure import urls as structure_urls
from waldur_core.users import urls as users_urls

CostTrackingRegister.autodiscover()

router = DefaultRouter()
cost_tracking_urls.register_in(router)
logging_urls.register_in(router)
monitoring_urls.register_in(router)
quotas_urls.register_in(router)
structure_urls.register_in(router)
users_urls.register_in(router)


urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(r'^admintools/', include('admin_tools.urls')),
    url(r'^admin/defender/', include('defender.urls')),
]

if settings.WALDUR_CORE.get('EXTENSIONS_AUTOREGISTER'):
    for ext in WaldurExtension.get_extensions():
        if ext.django_app() in settings.INSTALLED_APPS:
            urlpatterns += ext.django_urls()
            ext.rest_urls()(router)

urlpatterns += [
    url(r'^docs/', WaldurSchemaView.as_view()),
    url(r'^api/', include(router.urls)),
    url(r'^api/', include('waldur_core.logging.urls')),
    url(r'^api/', include('waldur_core.media.urls')),
    url(r'^api/', include('waldur_core.structure.urls')),
    url(r'^api/version/', core_views.version_detail),
    url(r'^api/configuration/', core_views.configuration_detail),
    url(r'^api-auth/password/', core_views.obtain_auth_token, name='auth-password'),
    url(r'^$', TemplateView.as_view(template_name='landing/index.html')),
]

if settings.DEBUG:
    from django.conf.urls.static import static
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

    # enable login/logout for web UI in debug mode
    urlpatterns += url(r'^api-auth/', include('rest_framework.urls', namespace='rest_framework')),



from __future__ import unicode_literals

import logging

from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers as rf_serializers, status
from rest_framework.decorators import detail_route
from rest_framework.mixins import RetrieveModelMixin
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet

from waldur_core.core import validators as core_validators
from waldur_core.structure import models as structure_models
from waldur_core.structure import views as structure_views
from waldur_vmware.apps import VMwareConfig

from . import filters, executors, models, serializers


logger = logging.getLogger(__name__)


class ServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.VMwareService.objects.all()
    serializer_class = serializers.ServiceSerializer


class ServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.VMwareServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer
    filter_class = filters.ServiceProjectLinkFilter


class LimitViewSet(RetrieveModelMixin, GenericViewSet):
    """
    Service consumer is not allowed to get details of service settings of service provider.
    However, currently VMware virtual machine limits are stored as options in service settings.
    Therefore in order to implement frontent-side validation of VM configuration in deployment form,
    we need to get limits of VMware service settings for service consumer.
    That's why GenericRoleFilter is not applied here.
    It is expected that eventually service provider limits would be moved to marketplace offering.
    """
    queryset = structure_models.ServiceSettings.objects.filter(type=VMwareConfig.service_name)
    lookup_field = 'uuid'
    serializer_class = serializers.LimitSerializer


class VirtualMachineViewSet(structure_views.BaseResourceViewSet):
    queryset = models.VirtualMachine.objects.all()
    serializer_class = serializers.VirtualMachineSerializer
    filter_class = filters.VirtualMachineFilter
    pull_executor = executors.VirtualMachinePullExecutor
    create_executor = executors.VirtualMachineCreateExecutor
    delete_executor = executors.VirtualMachineDeleteExecutor
    update_executor = executors.VirtualMachineUpdateExecutor
    update_validators = partial_update_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(models.VirtualMachine.RuntimeStates.POWERED_OFF),
    ]

    destroy_validators = structure_views.BaseResourceViewSet.destroy_validators + [
        core_validators.RuntimeStateValidator(models.VirtualMachine.RuntimeStates.POWERED_OFF)
    ]

    @detail_route(methods=['post'])
    def start(self, request, uuid=None):
        instance = self.get_object()
        executors.VirtualMachineStartExecutor().execute(instance)
        return Response({'status': _('start was scheduled')}, status=status.HTTP_202_ACCEPTED)

    start_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(
            models.VirtualMachine.RuntimeStates.POWERED_OFF,
            models.VirtualMachine.RuntimeStates.SUSPENDED,
        ),
    ]
    start_serializer_class = rf_serializers.Serializer

    @detail_route(methods=['post'])
    def stop(self, request, uuid=None):
        instance = self.get_object()
        executors.VirtualMachineStopExecutor().execute(instance)
        return Response({'status': _('stop was scheduled')}, status=status.HTTP_202_ACCEPTED)

    stop_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(
            models.VirtualMachine.RuntimeStates.POWERED_ON,
            models.VirtualMachine.RuntimeStates.SUSPENDED,
        ),
    ]
    stop_serializer_class = rf_serializers.Serializer

    @detail_route(methods=['post'])
    def reset(self, request, uuid=None):
        instance = self.get_object()
        executors.VirtualMachineResetExecutor().execute(instance)
        return Response({'status': _('reset was scheduled')}, status=status.HTTP_202_ACCEPTED)

    reset_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(
            models.VirtualMachine.RuntimeStates.POWERED_ON,
        ),
    ]
    reset_serializer_class = rf_serializers.Serializer

    @detail_route(methods=['post'])
    def suspend(self, request, uuid=None):
        instance = self.get_object()
        executors.VirtualMachineSuspendExecutor().execute(instance)
        return Response({'status': _('suspend was scheduled')}, status=status.HTTP_202_ACCEPTED)

    suspend_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(
            models.VirtualMachine.RuntimeStates.POWERED_ON,
        ),
    ]
    suspend_serializer_class = rf_serializers.Serializer

    def vm_tools_are_running(vm):
        if vm.tools_state != models.VirtualMachine.ToolsStates.RUNNING:
            raise rf_serializers.ValidationError('VMware Tools are not running.')

    @detail_route(methods=['post'])
    def shutdown_guest(self, request, uuid=None):
        instance = self.get_object()
        executors.VirtualMachineShutdownGuestExecutor().execute(instance)
        return Response({'status': _('shutdown was scheduled')}, status=status.HTTP_202_ACCEPTED)

    shutdown_guest_validators = reboot_guest_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(
            models.VirtualMachine.RuntimeStates.POWERED_ON,
        ),
        vm_tools_are_running,
    ]
    shutdown_guest_serializer_class = rf_serializers.Serializer

    @detail_route(methods=['post'])
    def reboot_guest(self, request, uuid=None):
        instance = self.get_object()
        executors.VirtualMachineRebootGuestExecutor().execute(instance)
        return Response({'status': _('reboot was scheduled')}, status=status.HTTP_202_ACCEPTED)

    reboot_guest_serializer_class = rf_serializers.Serializer

    @detail_route(methods=['post'])
    def create_port(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        port = serializer.save()

        transaction.on_commit(lambda: executors.PortCreateExecutor().execute(port))
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def check_number_of_ports(vm):
        # Limit of the network adapter per VM is 10 in vSphere 6.7, 6.5 and 6.0
        if vm.port_set.count() >= 10:
            raise rf_serializers.ValidationError('Virtual machine can have at most 10 network adapters.')

    create_port_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        check_number_of_ports,
    ]
    create_port_serializer_class = serializers.PortSerializer

    @detail_route(methods=['post'])
    def create_disk(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        disk = serializer.save()

        transaction.on_commit(lambda: executors.DiskCreateExecutor().execute(disk))
        return Response(serializer.data, status=status.HTTP_201_CREATED)

    def validate_total_size(vm):
        max_disk_total = serializers.get_int_or_none(vm.service_settings.options, 'max_disk_total')

        if max_disk_total:
            remaining_quota = max_disk_total - vm.total_disk
            if remaining_quota < 1024:
                raise rf_serializers.ValidationError('Storage quota has been reached.')

    create_disk_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        validate_total_size,
    ]
    create_disk_serializer_class = serializers.DiskSerializer

    @detail_route(methods=['get'])
    def console(self, request, uuid=None):
        """
        This endpoint provides access to Virtual Machine Remote Console aka VMRC.
        """
        instance = self.get_object()
        backend = instance.get_backend()
        try:
            url = backend.get_console_url(instance)
        except Exception:
            logger.exception('Unable to get console URL.')
            raise rf_serializers.ValidationError('Unable to get console URL.')
        return Response({'url': url}, status=status.HTTP_200_OK)

    console_validators = [core_validators.StateValidator(models.VirtualMachine.States.OK)]

    @detail_route(methods=['get'])
    def web_console(self, request, uuid=None):
        """
        This endpoint provides access to HTML Console aka WMKS.
        """
        instance = self.get_object()
        backend = instance.get_backend()
        try:
            url = backend.get_web_console_url(instance)
        except Exception:
            logger.exception('Unable to get web console URL.')
            raise rf_serializers.ValidationError('Unable to get web console URL.')
        return Response({'url': url}, status=status.HTTP_200_OK)

    web_console_validators = [
        core_validators.StateValidator(models.VirtualMachine.States.OK),
        core_validators.RuntimeStateValidator(models.VirtualMachine.RuntimeStates.POWERED_ON)
    ]


class PortViewSet(structure_views.BaseResourceViewSet):
    queryset = models.Port.objects.all().order_by('-created')
    serializer_class = serializers.PortSerializer
    filter_class = filters.PortFilter
    disabled_actions = ['create', 'update', 'partial_update']
    pull_executor = executors.PortPullExecutor
    delete_executor = executors.PortDeleteExecutor


class DiskViewSet(structure_views.BaseResourceViewSet):
    queryset = models.Disk.objects.all().order_by('-created')
    serializer_class = serializers.DiskSerializer
    filter_class = filters.DiskFilter
    disabled_actions = ['create', 'update', 'partial_update']
    pull_executor = executors.DiskPullExecutor
    delete_executor = executors.DiskDeleteExecutor

    @detail_route(methods=['post'])
    def extend(self, request, uuid=None):
        """ Increase disk capacity """
        disk = self.get_object()
        serializer = self.get_serializer(disk, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        disk.refresh_from_db()
        transaction.on_commit(lambda: executors.DiskExtendExecutor().execute(disk))

        return Response({'status': _('extend was scheduled')}, status=status.HTTP_202_ACCEPTED)

    def validate_total_size(disk):
        options = disk.vm.service_settings.options

        max_disk = serializers.get_int_or_none(options, 'max_disk')
        if max_disk and abs(max_disk - disk.size) < 1024:
            raise rf_serializers.ValidationError('Storage limit has been reached.')

        max_disk_total = serializers.get_int_or_none(options, 'max_disk_total')

        if max_disk_total:
            remaining_quota = max_disk_total - disk.vm.total_disk
            if remaining_quota < 1024:
                raise rf_serializers.ValidationError('Storage quota has been reached.')

    extend_validators = [core_validators.StateValidator(models.Disk.States.OK), validate_total_size]
    extend_serializer_class = serializers.DiskExtendSerializer


class TemplateViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Template.objects.all()
    serializer_class = serializers.TemplateSerializer
    filter_class = filters.TemplateFilter
    lookup_field = 'uuid'


class ClusterViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Cluster.objects.all()
    serializer_class = serializers.ClusterSerializer
    filter_class = filters.ClusterFilter
    lookup_field = 'uuid'


class NetworkViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Network.objects.all()
    serializer_class = serializers.NetworkSerializer
    filter_class = filters.NetworkFilter
    lookup_field = 'uuid'


class DatastoreViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Datastore.objects.all()
    serializer_class = serializers.DatastoreSerializer
    filter_class = filters.DatastoreFilter
    lookup_field = 'uuid'


class FolderViewSet(structure_views.BaseServicePropertyViewSet):
    queryset = models.Folder.objects.all()
    serializer_class = serializers.FolderSerializer
    filter_class = filters.FolderFilter
    lookup_field = 'uuid'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.contrib.contenttypes.models import ContentType


def cleanup_tenant_quotas(apps, schema_editor):
    Tenant = apps.get_model('openstack', 'Tenant')
    Quota = apps.get_model('quotas', 'Quota')
    quota_names = ['ram', 'volumes_size', 'subnet_count', 'floating_ip_count', 'storage', 'snapshots', 'instances',
                   'network_count', 'volumes', 'vcpu', 'security_group_rule_count', 'snapshots_size',
                   'security_group_count']
    tenant_content_type = ContentType.objects.get_for_model(Tenant)
    Quota.objects.filter(content_type_id=tenant_content_type.id).exclude(name__in=quota_names).delete()


def cleanup_openstackservice_quotas(apps, schema_editor):
    OpenStackService = apps.get_model('openstack', 'OpenStackService')
    Quota = apps.get_model('quotas', 'Quota')
    quota_names = ['tenant_count', 'ram', 'vcpu', 'floating_ip_count', 'storage', 'snapshots', 'instances', 'volumes',
                   'security_group_rule_count', 'security_group_count']
    service_content_type = ContentType.objects.get_for_model(OpenStackService)
    Quota.objects.filter(content_type_id=service_content_type.id).exclude(name__in=quota_names).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('openstack', '0030_subnet_dns_nameservers'),
    ]

    operations = [
        migrations.RunPython(cleanup_tenant_quotas, elidable=True),
        migrations.RunPython(cleanup_openstackservice_quotas, elidable=True),
    ]



import re

from django.core.validators import RegexValidator
from django.utils.translation import ugettext_lazy as _

ClusterNameValidator = RegexValidator(
    regex=re.compile(r"^[a-z0-9]([-a-z0-9])+[a-z0-9]$"),
    message=_('Name must consist of lower case alphanumeric characters or \'-\', '
              'and must start and end with an alphanumeric character')
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-05-21 09:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0017_project_action'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='runtime_state',
            field=models.CharField(blank=True, max_length=150, verbose_name='runtime state'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0035_settings_tags_and_scope'),
        ('openstack', '0004_dr_and_volume_backups'),
    ]

    operations = [
        migrations.CreateModel(
            name='IpMapping',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('public_ip', models.GenericIPAddressField(protocol='IPv4')),
                ('private_ip', models.GenericIPAddressField(protocol='IPv4')),
                ('project', models.ForeignKey(related_name='+', to='structure.Project')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from celery import shared_task

from waldur_core.cost_tracking import CostTrackingRegister, models
from waldur_core.structure import models as structure_models


@shared_task(name='waldur_core.cost_tracking.recalculate_estimate')
def recalculate_estimate(recalculate_total=False):
    """ Recalculate price of consumables that were used by resource until now.

        Regular task. It is too expensive to calculate consumed price on each
        request, so we store cached price each hour.
        If recalculate_total is True - task also recalculates total estimate
        for current month.
    """
    # Celery does not import server.urls and does not discover cost tracking modules.
    # So they should be discovered implicitly.
    CostTrackingRegister.autodiscover()
    # Step 1. Recalculate resources estimates.
    for resource_model in CostTrackingRegister.registered_resources:
        for resource in resource_model.objects.all():
            _update_resource_consumed(resource, recalculate_total=recalculate_total)
    # Step 2. Move from down to top and recalculate consumed estimate for each
    #         object based on its children.
    ancestors_models = [m for m in models.PriceEstimate.get_estimated_models()
                        if not issubclass(m, structure_models.ResourceMixin)]
    for model in ancestors_models:
        for ancestor in model.objects.all():
            _update_ancestor_consumed(ancestor)


def _update_resource_consumed(resource, recalculate_total):
    price_estimate, created = models.PriceEstimate.objects.get_or_create_current(scope=resource)
    if created:
        models.ConsumptionDetails.objects.create(price_estimate=price_estimate)
        price_estimate.create_ancestors()
        price_estimate.update_total()
    elif recalculate_total:
        price_estimate.update_total()
    price_estimate.update_consumed()


def _update_ancestor_consumed(ancestor):
    price_estimate, _ = models.PriceEstimate.objects.get_or_create_current(scope=ancestor)
    resource_descendants = [descendant for descendant in price_estimate.get_descendants()
                            if isinstance(descendant.scope, structure_models.ResourceMixin)]
    price_estimate.consumed = sum([descendant.consumed for descendant in resource_descendants])
    price_estimate.save(update_fields=['consumed'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from waldur_openstack.openstack_tenant.models import Backup


def add_backup_size_to_metadata(apps, schema_editor):
    for backup in Backup.objects.iterator():
        backup.metadata['size'] = backup.instance.size
        backup.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0023_remove_instance_external_ip'),
    ]

    operations = [
        migrations.RunPython(add_backup_size_to_metadata, elidable=True),
    ]



from __future__ import unicode_literals

import logging
import re

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import JSONField as BetterJSONField
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from django_fsm import FSMIntegerField
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel

from waldur_core.core import fields as core_fields
from waldur_core.core import models as core_models
from waldur_core.core.validators import validate_name
from waldur_core.structure import models as structure_models
from waldur_mastermind.common import mixins as common_mixins

from . import backend, managers

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class Issue(core_models.UuidMixin,
            structure_models.StructureLoggableMixin,
            core_models.BackendModelMixin,
            TimeStampedModel,
            core_models.StateMixin):
    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'customer'
        project_path = 'project'

    backend_id = models.CharField(max_length=255, blank=True, null=True, unique=True)
    key = models.CharField(max_length=255, blank=True)
    type = models.CharField(max_length=255)
    link = models.URLField(max_length=255, help_text=_('Link to issue in support system.'), blank=True)

    summary = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    deadline = models.DateTimeField(blank=True, null=True)
    impact = models.CharField(max_length=255, blank=True)

    status = models.CharField(max_length=255)
    resolution = models.CharField(max_length=255, blank=True)
    priority = models.CharField(max_length=255, blank=True)

    caller = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='created_issues',
                               help_text=_('Waldur user who has reported the issue.'),
                               on_delete=models.PROTECT)
    reporter = models.ForeignKey('SupportUser', related_name='reported_issues', blank=True, null=True,
                                 help_text=_('Help desk user who have created the issue that is reported by caller.'),
                                 on_delete=models.PROTECT)
    assignee = models.ForeignKey('SupportUser', related_name='issues', blank=True, null=True,
                                 help_text=_('Help desk user who will implement the issue'),
                                 on_delete=models.PROTECT)

    customer = models.ForeignKey(structure_models.Customer, verbose_name=_('organization'),
                                 related_name='issues', blank=True, null=True,
                                 on_delete=models.CASCADE)
    project = models.ForeignKey(structure_models.Project, related_name='issues', blank=True, null=True,
                                on_delete=models.CASCADE)

    resource_content_type = models.ForeignKey(ContentType, null=True)
    resource_object_id = models.PositiveIntegerField(null=True)
    resource = GenericForeignKey('resource_content_type', 'resource_object_id')

    first_response_sla = models.DateTimeField(blank=True, null=True)
    resolution_date = models.DateTimeField(blank=True, null=True)
    template = models.ForeignKey('Template', related_name='issues', blank=True, null=True, on_delete=models.PROTECT)

    tracker = FieldTracker()

    def get_description(self):
        return self.description

    @classmethod
    def get_url_name(cls):
        return 'support-issue'

    @classmethod
    def get_backend_fields(cls):
        return super(Issue, cls).get_backend_fields() + ('backend_id', 'key', 'type', 'link',
                                                         'summary', 'description', 'deadline', 'impact',
                                                         'status', 'resolution', 'priority',
                                                         'caller', 'reporter', 'assignee', 'customer', 'project',
                                                         'resource', 'first_response_sla')

    def get_log_fields(self):
        return ('uuid', 'type', 'key', 'status', 'link', 'summary',
                'reporter', 'caller', 'customer', 'project', 'resource')

    @property
    def resolved(self):
        return IssueStatus.check_success_status(self.status)

    def set_resolved(self):
        self.status = IssueStatus.objects.filter(type=IssueStatus.Types.RESOLVED).first().name
        self.state = Issue.States.OK
        self.save()

    def set_canceled(self):
        self.status = IssueStatus.objects.filter(type=IssueStatus.Types.CANCELED).first().name
        self.state = Issue.States.OK
        self.save()

    def __str__(self):
        return '{}: {}'.format(self.key or '???', self.summary)


@python_2_unicode_compatible
class Priority(core_models.NameMixin,
               core_models.UuidMixin,
               core_models.UiDescribableMixin):
    backend_id = models.CharField(max_length=255, blank=True)

    class Meta(object):
        verbose_name = _('Priority')
        verbose_name_plural = _('Priorities')

    @classmethod
    def get_url_name(cls):
        return 'support-priority'

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class SupportUser(core_models.UuidMixin, core_models.NameMixin, models.Model):
    class Meta:
        ordering = ['name']

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='+', blank=True, null=True)
    backend_id = models.CharField(max_length=255, blank=True, null=True, unique=True)
    is_active = models.BooleanField(_('active'), default=True,
                                    help_text=_('Designates whether this user should be treated as '
                                                'active. Unselect this instead of deleting accounts.'))
    objects = managers.SupportUserManager()

    @classmethod
    def get_url_name(cls):
        return 'support-user'

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class Comment(core_models.UuidMixin,
              core_models.BackendModelMixin,
              TimeStampedModel,
              core_models.StateMixin):
    class Meta:
        ordering = ['-created']
        unique_together = ('backend_id', 'issue')

    class Permissions(object):
        customer_path = 'issue__customer'
        project_path = 'issue__project'

    issue = models.ForeignKey(Issue, related_name='comments')
    author = models.ForeignKey(SupportUser, related_name='comments')
    description = models.TextField()
    is_public = models.BooleanField(default=True)
    backend_id = models.CharField(max_length=255, blank=True, null=True)

    def clean_message(self, message):
        """
        Extracts comment message from JIRA comment which contains user's info in its body.
        """
        match = re.search('^(\[.*?\]\:\s)', message)
        return message.replace(match.group(0), '') if match else message

    def prepare_message(self):
        """
        Prepends user info to the comment description to display comment author in JIRA.
        User info format - '[user.full_name user.civil_number]: '.
        """
        prefix = self.author.name
        # User is optional
        user = self.author.user
        if user:
            prefix = user.full_name or user.username
            if user.civil_number:
                prefix += ' ' + user.civil_number
        return '[%s]: %s' % (prefix, self.description)

    def update_message(self, message):
        self.description = self.clean_message(message)

    @classmethod
    def get_url_name(cls):
        return 'support-comment'

    @classmethod
    def get_backend_fields(cls):
        return super(Comment, cls).get_backend_fields() + ('issue', 'author', 'description', 'is_public', 'backend_id')

    def __str__(self):
        return self.description[:50]


@python_2_unicode_compatible
class Offering(core_models.UuidMixin,
               core_models.NameMixin,
               common_mixins.ProductCodeMixin,
               common_mixins.UnitPriceMixin,
               structure_models.StructureLoggableMixin,
               TimeStampedModel):

    class Meta:
        ordering = ['-created']
        verbose_name = _('Request')
        verbose_name_plural = _('Requests')

    class Permissions(object):
        customer_path = 'project__customer'
        project_path = 'project'

    class States(object):
        REQUESTED = 'requested'
        OK = 'ok'
        TERMINATED = 'terminated'

        CHOICES = ((REQUESTED, _('Requested')), (OK, _('OK')), (TERMINATED, _('Terminated')))

    template = models.ForeignKey('OfferingTemplate', on_delete=models.PROTECT)
    plan = models.ForeignKey('OfferingPlan', blank=True, null=True, on_delete=models.PROTECT)
    issue = models.ForeignKey(Issue, null=True, on_delete=models.PROTECT)
    project = models.ForeignKey(structure_models.Project, null=True, on_delete=models.PROTECT)
    state = models.CharField(default=States.REQUESTED, max_length=30, choices=States.CHOICES)
    report = core_fields.JSONField(blank=True)
    terminated_at = models.DateTimeField(editable=False, blank=True, null=True)

    tracker = FieldTracker()

    def get_backend(self):
        backend.get_active_backend()

    def get_log_fields(self):
        return super(Offering, self).get_log_fields() + ('state', )

    def terminate(self):
        self.state = Offering.States.TERMINATED
        self.save()

    def set_ok(self):
        self.state = Offering.States.OK
        self.save()

    @property
    def type(self):
        return self.template.name

    @property
    def type_label(self):
        return self.template.config.get('label', None)

    @classmethod
    def get_url_name(cls):
        return 'support-offering'

    def __str__(self):
        return '{}: {}'.format(self.type_label or self.name, self.state)

    @classmethod
    def get_scope_type(cls):
        return 'Support.Offering'

    def _get_log_context(self, entity_name):
        context = super(Offering, self)._get_log_context(entity_name)
        context['resource_type'] = self.get_scope_type()
        context['resource_uuid'] = self.uuid.hex
        return context

    @property
    def config(self):
        return self.template.config if self.template else {}


@python_2_unicode_compatible
class OfferingTemplate(core_models.UuidMixin,
                       TimeStampedModel):
    name = models.CharField(_('name'), max_length=150)
    config = BetterJSONField()
    sort_order = models.PositiveSmallIntegerField(default=1)

    class Meta:
        ordering = ['sort_order', 'name']

    @classmethod
    def get_url_name(cls):
        return 'support-offering-template'

    def __str__(self):
        return self.name


class OfferingPlan(core_models.UuidMixin,
                   core_models.NameMixin,
                   core_models.DescribableMixin,
                   common_mixins.ProductCodeMixin,
                   common_mixins.UnitPriceMixin):
    template = models.ForeignKey(OfferingTemplate, related_name='plans')

    def __str__(self):
        return '{} | {}'.format(self.template, self.name)


@python_2_unicode_compatible
class Attachment(core_models.UuidMixin,
                 TimeStampedModel,
                 structure_models.StructureLoggableMixin,
                 core_models.StateMixin):
    class Permissions(object):
        customer_path = 'issue__customer'
        project_path = 'issue__project'

    issue = models.ForeignKey(Issue, related_name='attachments')
    file = models.FileField(upload_to='support_attachments')
    backend_id = models.CharField(max_length=255, unique=True)
    mime_type = models.CharField(_('MIME type'), max_length=100, blank=True)
    file_size = models.PositiveIntegerField(_('Filesize, B'), blank=True, null=True)
    thumbnail = models.FileField(upload_to='support_attachments_thumbnails', blank=True, null=True)
    author = models.ForeignKey(SupportUser, related_name='attachments', blank=True, null=True)
    objects = managers.AttachmentManager()

    @classmethod
    def get_url_name(cls):
        return 'support-attachment'

    def __str__(self):
        return '{} | {}'.format(self.issue, self.file.name.split('/')[-1])

    def get_log_fields(self):
        return ('uuid', 'issue', 'author', 'backend_id')


class Template(core_models.UuidMixin,
               core_models.NameMixin,
               TimeStampedModel):

    class IssueTypes(object):
        INFORMATIONAL = 'INFORMATIONAL'
        SERVICE_REQUEST = 'SERVICE_REQUEST'
        CHANGE_REQUEST = 'CHANGE_REQUEST'
        INCIDENT = 'INCIDENT'

        CHOICES = (
            (INFORMATIONAL, 'Informational'),
            (SERVICE_REQUEST, 'Service request'),
            (CHANGE_REQUEST, 'Change request'),
            (INCIDENT, 'Incident'),
        )

    native_name = models.CharField(max_length=150, blank=True)
    description = models.TextField()
    native_description = models.TextField(blank=True)
    issue_type = models.CharField(max_length=30, choices=IssueTypes.CHOICES, default=IssueTypes.INFORMATIONAL)

    @classmethod
    def get_url_name(cls):
        return 'support-template'

    def __str__(self):
        return self.name


class TemplateAttachment(core_models.UuidMixin,
                         core_models.NameMixin,
                         TimeStampedModel):
    template = models.ForeignKey(Template, on_delete=models.CASCADE, related_name='attachments')
    file = models.FileField(upload_to='support_template_attachments')


class IgnoredIssueStatus(models.Model):
    name = models.CharField(_('name'), max_length=150, validators=[validate_name], unique=True)

    def __str__(self):
        return self.name


class TemplateStatusNotification(models.Model):
    status = models.CharField(max_length=255, validators=[validate_name], unique=True)
    html = models.TextField(validators=[validate_name])
    text = models.TextField(validators=[validate_name])
    subject = models.CharField(max_length=255, validators=[validate_name])

    def __str__(self):
        return self.status


class SupportCustomer(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    backend_id = models.CharField(max_length=255, unique=True)


class RequestType(core_models.UuidMixin, core_models.NameMixin, models.Model):
    backend_id = models.IntegerField(unique=True)
    issue_type_name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


class IssueStatus(models.Model):
    """ This model is needed in order to understand whether the issue has been solved or not.

        The field of resolution does not give an exact answer since may be the same in both cases.
    """

    class Types(object):
        RESOLVED = 0
        CANCELED = 1

    TYPE_CHOICES = (
        (Types.RESOLVED, 'Resolved'),
        (Types.CANCELED, 'Canceled'),
    )

    name = models.CharField(max_length=255, help_text='Status name in Jira.', unique=True)
    type = FSMIntegerField(default=Types.RESOLVED, choices=TYPE_CHOICES)

    @classmethod
    def check_success_status(cls, status):
        """ Check an issue has been resolved.

            True if an issue resolved.
            False if an issue canceled.
            None in all other cases.
        """
        if not cls.objects.filter(type=cls.Types.RESOLVED).exists() or \
                not cls.objects.filter(type=cls.Types.CANCELED).exists():
            logger.critical('There is no information about statuses of an issue. '
                            'Please, add resolved and cancelled statuses in admin. '
                            'Otherwise, you cannot use processes that need this information.')
            return
        try:
            issue_status = cls.objects.get(name=status)
            if issue_status.type == cls.Types.RESOLVED:
                return True
            if issue_status.type == cls.Types.CANCELED:
                return False
        except cls.DoesNotExist:
            return

    class Meta:
        verbose_name = _('Issue status')
        verbose_name_plural = _('Issue statuses')



from __future__ import unicode_literals

from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import admin as core_admin
from waldur_core.core.admin import JsonWidget
from waldur_core.structure import admin as structure_admin

from . import models, backend
from .backend.basic import BasicBackend

User = get_user_model()


class UserChoiceField(forms.ModelChoiceField):
    def label_from_instance(self, user):
        return '{} - {}'.format(user.full_name, user.username)


class SupportUserAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(SupportUserAdminForm, self).__init__(*args, **kwargs)
        self.fields['user'] = UserChoiceField(queryset=User.objects.all().order_by('full_name'))


class SupportUserAdmin(admin.ModelAdmin):
    form = SupportUserAdminForm


class OfferingAdminForm(forms.ModelForm):
    class Meta:
        widgets = {
            'report': JsonWidget(),
        }


class OfferingAdmin(admin.ModelAdmin):
    list_display = ('template', 'name', 'unit_price', 'unit', 'state',
                    'created', 'modified', 'issue_key')
    search_fields = ('name', 'template__name', 'issue__key')
    fields = ('name', 'unit_price', 'unit', 'template', 'issue',
              'project', 'state', 'product_code', 'article_code', 'report')
    form = OfferingAdminForm

    def issue_key(self, offering):
        return offering.issue and offering.issue.key or 'N/A'


class OfferingTemplateAdminForm(forms.ModelForm):
    class Meta:
        widgets = {
            'config': JsonWidget(),
        }


class OfferingTemplateAdmin(admin.ModelAdmin):
    form = OfferingTemplateAdminForm


class IssueAdmin(core_admin.ExtraActionsObjectMixin, structure_admin.BackendModelAdmin):
    exclude = ('resource_content_type', 'resource_object_id')
    ordering = ('-created',)
    search_fields = ('key', 'backend_id', 'summary')
    list_filter = ('type', 'status', 'resolution')
    list_display = ('key', 'summary', 'type', 'status', 'resolution', 'get_caller_full_name')

    def get_caller_full_name(self, obj):
        if obj.caller:
            return obj.caller.full_name
        return

    get_caller_full_name.short_description = 'Caller name'

    def resolve(self, request, pk=None):
        issue = get_object_or_404(models.Issue, pk=pk)
        issue.set_resolved()
        message = _('Issue has been resolved.')
        self.message_user(request, message)
        return HttpResponseRedirect('../')

    def cancel(self, request, pk=None):
        issue = get_object_or_404(models.Issue, pk=pk)
        issue.set_canceled()
        message = _('Issue has been canceled.')
        self.message_user(request, message)
        return HttpResponseRedirect('../')

    def buttons_validate(request, obj):
        if isinstance(backend.get_active_backend(), BasicBackend) and obj.resolved is None:
            return True

    resolve.validator = buttons_validate
    cancel.validator = buttons_validate

    def get_extra_object_actions(self):
        return [
            self.resolve,
            self.cancel,
        ]


class TemplateAttachmentInline(admin.TabularInline):
    model = models.TemplateAttachment
    fields = ('name', 'file')


class TemplateAdmin(core_admin.ExcludedFieldsAdminMixin,
                    admin.ModelAdmin):
    list_display = ('name', 'issue_type', 'created')
    search_fields = ('name', 'native_name')
    fields = ('name', 'native_name',
              'description', 'native_description',
              'issue_type', 'created', 'modified')
    readonly_fields = ('created', 'modified')
    inlines = [TemplateAttachmentInline]

    @cached_property
    def excluded_fields(self):
        if not settings.WALDUR_CORE['NATIVE_NAME_ENABLED']:
            return ['native_name', 'native_description']
        return []


class RequestTypeAdmin(admin.ModelAdmin):
    list_display = ('name', 'issue_type_name', 'backend_id')
    search_fields = ('name',)


class PriorityAdmin(admin.ModelAdmin):
    list_display = ('name', 'description', 'backend_id')
    search_fields = ('name', 'description')


class IssueStatusAdmin(admin.ModelAdmin):
    list_display = ('name', 'type')


class CommentAdmin(structure_admin.BackendModelAdmin):
    list_display = ('get_issue_key', 'is_public', 'author', 'created')
    list_filter = ('is_public', 'author')
    search_fields = ('description',)

    def get_issue_key(self, obj):
        return "%s: %s" % (obj.issue.key, obj.issue.summary)

    get_issue_key.short_description = 'Issue'


admin.site.register(models.Offering, OfferingAdmin)
admin.site.register(models.Issue, IssueAdmin)
admin.site.register(models.Comment, CommentAdmin)
admin.site.register(models.Attachment)
admin.site.register(models.SupportUser, SupportUserAdmin)
admin.site.register(models.Template, TemplateAdmin)
admin.site.register(models.OfferingTemplate, OfferingTemplateAdmin)
admin.site.register(models.OfferingPlan)
admin.site.register(models.TemplateStatusNotification)
admin.site.register(models.IgnoredIssueStatus)
admin.site.register(models.RequestType, RequestTypeAdmin)
admin.site.register(models.Priority, PriorityAdmin)
admin.site.register(models.IssueStatus, IssueStatusAdmin)



from __future__ import unicode_literals

import copy
import logging
import re

from django.conf import settings
from django.core import validators
from django.contrib.auth import password_validation
from django.db import transaction
from django.db.models import Q
from django.template.defaultfilters import slugify
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from iptools.ipv4 import validate_cidr

from waldur_core.core import utils as core_utils, serializers as core_serializers
from waldur_core.quotas import serializers as quotas_serializers
from waldur_core.structure import serializers as structure_serializers, permissions as structure_permissions

from . import models
from .backend import OpenStackBackendError


logger = logging.getLogger(__name__)


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        core_serializers.RequiredFieldsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': _('Keystone auth URL (e.g. http://keystone.example.com:5000/v3)'),
        'username': _('Administrative user'),
        'domain': _('Domain name. If not defined default domain will be used.'),
        'password': '',
    }
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'tenant_name': '',
        'availability_zone': _('Default availability zone for provisioned instances'),
        'volume_availability_zone_name': _('Default availability zone name for provisioned volumes'),
        'valid_availability_zones': _('Optional dictionary where key is Nova availability '
                                      'zone name and value is Cinder availability zone name.'),
        'external_network_id': _('ID of OpenStack external network that will be connected to tenants'),
        'latitude': _('Latitude of the datacenter (e.g. 40.712784)'),
        'longitude': _('Longitude of the datacenter (e.g. -74.005941)'),
        'access_url': _('Publicly accessible OpenStack dashboard URL'),
        'dns_nameservers': _('Default value for new subnets DNS name servers. Should be defined as list.'),
        'flavor_exclude_regex': _('Flavors matching this regex expression will not be pulled from the backend.'),
        'create_ha_routers': _('Create highly available Neutron routers.'),
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.OpenStackService
        required_fields = 'backend_url', 'username', 'password', 'console_type'
        extra_field_options = {
            'backend_url': {
                'label': 'API URL',
                'default_value': 'http://keystone.example.com:5000/v2.0',
            },
            'username': {
                'default_value': 'admin',
            },
            'tenant_name': {
                'label': 'Tenant name',
                'default_value': 'admin',
            },
            'external_network_id': {
                'label': 'Public/gateway network UUID',
            },
            'availability_zone': {
                'placeholder': 'default',
            },
            'volume_availability_zone_name': {
                'label': 'Name of default volume availability zone to use',
            },
            'access_url': {
                'label': 'Access URL',
            },
            'create_ha_routers': {
                'default_value': False,
            },
            'console_type': {
                'default_value': 'novnc',
            },
        }

    def _validate_settings(self, settings):
        backend = settings.get_backend()
        try:
            if not backend.check_admin_tenant():
                raise serializers.ValidationError({
                    'non_field_errors': _('Provided credentials are not for admin tenant.')
                })
        except OpenStackBackendError:
            raise serializers.ValidationError({
                'non_field_errors': _('Unable to validate credentials.')
            })


class ServiceNameSerializer(serializers.Serializer):
    name = serializers.CharField(required=True)


class FlavorSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Flavor
        fields = ('url', 'uuid', 'name', 'cores', 'ram', 'disk', 'display_name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    display_name = serializers.SerializerMethodField()

    def get_display_name(self, flavor):
        return "{} ({} CPU, {} MB RAM, {} MB HDD)".format(
            flavor.name, flavor.cores, flavor.ram, flavor.disk)


class ImageSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Image
        fields = ('url', 'uuid', 'name', 'min_disk', 'min_ram')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):

    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.OpenStackServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'openstack-detail'},
        }


class TenantQuotaSerializer(serializers.Serializer):
    instances = serializers.IntegerField(min_value=1, required=False)
    volumes = serializers.IntegerField(min_value=1, required=False)
    snapshots = serializers.IntegerField(min_value=1, required=False)
    ram = serializers.IntegerField(min_value=1, required=False)
    vcpu = serializers.IntegerField(min_value=1, required=False)
    storage = serializers.IntegerField(min_value=1, required=False)
    security_group_count = serializers.IntegerField(min_value=1, required=False)
    security_group_rule_count = serializers.IntegerField(min_value=1, required=False)


class FloatingIPSerializer(structure_serializers.BaseResourceActionSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstack-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        read_only=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.FloatingIP
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'runtime_state', 'address', 'backend_network_id',
            'tenant', 'tenant_name', 'tenant_uuid')
        related_paths = ('tenant',)
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'runtime_state', 'address', 'description', 'name', 'tenant', 'backend_network_id',
            'service_settings', 'project')
        extra_kwargs = dict(
            tenant={'lookup_field': 'uuid', 'view_name': 'openstack-tenant-detail'},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        attrs['tenant'] = tenant = self.context['view'].get_object()
        attrs['service_project_link'] = tenant.service_project_link
        return super(FloatingIPSerializer, self).validate(attrs)


class SecurityGroupRuleSerializer(serializers.ModelSerializer):

    class Meta:
        model = models.SecurityGroupRule
        fields = ('id', 'protocol', 'from_port', 'to_port', 'cidr')

    def validate(self, rule):
        """
        Please note that validate function accepts rule object instead of validated data
        because it is used as a child of list serializer.
        """
        protocol = rule.protocol
        from_port = rule.from_port
        to_port = rule.to_port
        cidr = rule.cidr

        if cidr and not validate_cidr(cidr):
            raise serializers.ValidationError({
                'cidr': _('Expected cidr format: <0-255>.<0-255>.<0-255>.<0-255>/<0-32>')
            })

        if to_port is None:
            raise serializers.ValidationError({
                'to_port': _('Empty value is not allowed.')
            })

        if from_port is None:
            raise serializers.ValidationError({
                'from_port': _('Empty value is not allowed.')
            })

        if protocol == 'icmp':
            if from_port is not None and not -1 <= from_port <= 255:
                raise serializers.ValidationError({
                    'from_port': _('Value should be in range [-1, 255], found %d') % from_port})
            if to_port is not None and not -1 <= to_port <= 255:
                raise serializers.ValidationError({
                    'to_port': _('Value should be in range [-1, 255], found %d') % to_port
                })

        elif protocol in ('tcp', 'udp'):
            if from_port is not None and to_port is not None:
                if from_port > to_port:
                    raise serializers.ValidationError(_('"from_port" should be less or equal to "to_port"'))
            if from_port is not None and from_port < 1:
                raise serializers.ValidationError({
                    'from_port': _('Value should be in range [1, 65535], found %d') % from_port
                })
            if to_port is not None and to_port < 1:
                raise serializers.ValidationError({
                    'to_port': _('Value should be in range [1, 65535], found %d') % to_port
                })

        elif protocol == '':
            # See also: https://github.com/openstack/neutron/blob/af130e79cbe5d12b7c9f9f4dcbcdc8d972bfcfd4/neutron/db/securitygroups_db.py#L500

            if from_port != -1:
                raise serializers.ValidationError({
                    'from_port': _('Port range is not supported if protocol is not specified.')
                })

            if to_port != -1:
                raise serializers.ValidationError({
                    'to_port': _('Port range is not supported if protocol is not specified.')
                })

        else:
            raise serializers.ValidationError({
                'protocol': _('Value should be one of (tcp, udp, icmp), found %s') % protocol
            })

        return rule


class SecurityGroupRuleCreateSerializer(SecurityGroupRuleSerializer):
    """ Create rules on security group creation """

    def to_internal_value(self, data):
        if 'id' in data:
            raise serializers.ValidationError(
                _('Cannot add existed rule with id %s to new security group') % data['id'])
        internal_data = super(SecurityGroupRuleSerializer, self).to_internal_value(data)
        return models.SecurityGroupRule(**internal_data)


class SecurityGroupRuleUpdateSerializer(SecurityGroupRuleSerializer):

    def to_internal_value(self, data):
        """ Create new rule if id is not specified, update exist rule if id is specified """
        security_group = self.context['view'].get_object()
        internal_data = super(SecurityGroupRuleSerializer, self).to_internal_value(data)
        if 'id' not in data:
            return models.SecurityGroupRule(security_group=security_group, **internal_data)
        rule_id = data.pop('id')
        try:
            rule = security_group.rules.get(id=rule_id)
        except models.SecurityGroupRule.DoesNotExist:
            raise serializers.ValidationError({'id': _('Security group does not have rule with id %s.') % rule_id})
        for key, value in internal_data.items():
            setattr(rule, key, value)
        return rule


def validate_duplicate_security_group_rules(rules):
    values = rules.values_list('protocol', 'from_port', 'to_port', 'cidr')
    if len(set(values)) != len(values):
        raise serializers.ValidationError(_('Duplicate security group rules are not allowed.'))


class SecurityGroupRuleListUpdateSerializer(serializers.ListSerializer):
    child = SecurityGroupRuleUpdateSerializer()

    @transaction.atomic()
    def save(self, **kwargs):
        security_group = self.context['view'].get_object()
        old_rules_count = security_group.rules.count()
        rules = self.validated_data
        security_group.rules.exclude(id__in=[r.id for r in rules if r.id]).delete()
        for rule in rules:
            rule.save()
        validate_duplicate_security_group_rules(security_group.rules)
        security_group.change_backend_quotas_usage_on_rules_update(old_rules_count)
        return rules


class SecurityGroupSerializer(structure_serializers.BaseResourceActionSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstack-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        read_only=True)
    rules = SecurityGroupRuleCreateSerializer(many=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.SecurityGroup
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'tenant', 'tenant_name', 'tenant_uuid', 'rules',
        )
        related_paths = ('tenant',)
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'service_settings', 'project')
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + ('rules',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'openstack-sgp-detail'},
            'tenant': {'lookup_field': 'uuid', 'view_name': 'openstack-tenant-detail', 'read_only': True},
        }

    def validate_rules(self, value):
        for rule in value:
            if rule.id is not None:
                raise serializers.ValidationError(
                    _('Cannot add existed rule with id %s to new security group') % rule.id)
            rule.full_clean(exclude=['security_group'])
        return value

    def validate_name(self, value):
        if value == 'default':
            raise serializers.ValidationError(_('Default security group is managed by OpenStack itself.'))
        return value

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        attrs['tenant'] = tenant = self.context['view'].get_object()
        attrs['service_project_link'] = tenant.service_project_link
        return super(SecurityGroupSerializer, self).validate(attrs)

    def create(self, validated_data):
        rules = validated_data.pop('rules', [])
        with transaction.atomic():
            # quota usage has to be increased only after rules creation,
            # so we cannot execute BaseResourceSerializer create method.
            security_group = super(structure_serializers.BaseResourceSerializer, self).create(validated_data)
            for rule in rules:
                security_group.rules.add(rule, bulk=False)
            validate_duplicate_security_group_rules(security_group.rules)
            security_group.increase_backend_quotas_usage()
        return security_group


class TenantImportableSerializer(serializers.Serializer):
    backend_id = serializers.CharField(read_only=True)
    name = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    type = serializers.CharField(read_only=True)
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        queryset=models.OpenStackServiceProjectLink.objects.all(),
        write_only=True)


class TenantImportSerializer(serializers.HyperlinkedModelSerializer):
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        write_only=True,
        queryset=models.OpenStackServiceProjectLink.objects.all()
    )
    quotas = quotas_serializers.QuotaSerializer(many=True, read_only=True)

    class Meta(object):
        model = models.Tenant
        read_only_fields = ('name', 'uuid', 'availability_zone', 'internal_network_id', 'external_network_id',
                            'user_username', 'user_password', 'quotas')
        fields = read_only_fields + ('service_project_link', 'backend_id')

    @transaction.atomic
    def create(self, validated_data):
        service_project_link = validated_data['service_project_link']
        backend = service_project_link.service.get_backend()
        backend_id = validated_data['backend_id']

        if models.Tenant.objects.filter(
                service_project_link__service__settings=service_project_link.service.settings,
                backend_id=backend_id).exists():
            raise serializers.ValidationError(_('Tenant with ID "%s" is already registered.') % backend_id)

        try:
            tenant = backend.import_tenant(backend_id, service_project_link)
        except OpenStackBackendError as e:
            raise serializers.ValidationError({
                'backend_id': _('Can\'t import tenant with ID %(backend_id)s. Reason: %(reason)s') % {
                    'backend_id': backend_id,
                    'reason': e,
                }
            })

        tenant.user_username = models.Tenant.generate_username(tenant.name)
        tenant.user_password = core_utils.pwgen()
        tenant.save()

        return tenant


subnet_cidr_validator = validators.RegexValidator(
    re.compile(settings.WALDUR_OPENSTACK['SUBNET']['CIDR_REGEX']),
    settings.WALDUR_OPENSTACK['SUBNET']['CIDR_REGEX_EXPLANATION'],
)


class TenantSerializer(structure_serializers.PrivateCloudSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstack-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        queryset=models.OpenStackServiceProjectLink.objects.all(),
        write_only=True)
    quotas = quotas_serializers.QuotaSerializer(many=True, read_only=True)
    subnet_cidr = serializers.CharField(
        validators=[subnet_cidr_validator], default='192.168.42.0/24', initial='192.168.42.0/24', write_only=True)

    class Meta(structure_serializers.PrivateCloudSerializer.Meta):
        model = models.Tenant
        fields = structure_serializers.PrivateCloudSerializer.Meta.fields + (
            'availability_zone', 'internal_network_id', 'external_network_id',
            'user_username', 'user_password', 'quotas', 'subnet_cidr', 'default_volume_type_name',
        )
        read_only_fields = structure_serializers.PrivateCloudSerializer.Meta.read_only_fields + (
            'internal_network_id', 'external_network_id',
        )
        protected_fields = structure_serializers.PrivateCloudSerializer.Meta.protected_fields + (
            'user_username', 'subnet_cidr', 'user_password',
        )

    def get_fields(self):
        fields = super(TenantSerializer, self).get_fields()
        if not settings.WALDUR_OPENSTACK['TENANT_CREDENTIALS_VISIBLE']:
            for field in ('user_username', 'user_password', 'access_url'):
                if field in fields:
                    del fields[field]

        return fields

    def _validate_service_project_link(self, spl):
        """ Administrator can create tenant only using not shared service settings """
        user = self.context['request'].user
        message = _('You do not have permissions to create tenant in this project using selected service.')
        if spl.service.settings.shared and not user.is_staff:
            raise serializers.ValidationError(message)
        if not spl.service.settings.shared and not structure_permissions._has_admin_access(user, spl.project):
            raise serializers.ValidationError(message)
        return spl

    def validate_security_groups_configuration(self):
        nc_settings = getattr(settings, 'WALDUR_OPENSTACK', {})
        config_groups = nc_settings.get('DEFAULT_SECURITY_GROUPS', [])
        for group in config_groups:
            sg_name = group.get('name')
            if sg_name in (None, ''):
                raise serializers.ValidationError(
                    _('Skipping misconfigured security group: parameter "name" not found or is empty.'))

            rules = group.get('rules')
            if type(rules) not in (list, tuple):
                raise serializers.ValidationError(
                    _('Skipping misconfigured security group: parameter "rules" should be list or tuple.'))

    def _get_neighbour_tenants(self, service_settings):
        domain = service_settings.domain
        backend_url = service_settings.backend_url
        tenants = models.Tenant.objects.filter(service_project_link__service__settings__backend_url=backend_url)
        if domain in (None, '', 'default'):
            tenants = tenants.filter(
                Q(service_project_link__service__settings__domain='') |
                Q(service_project_link__service__settings__domain__isnull=True) |
                Q(service_project_link__service__settings__domain__iexact='default')
            )
        else:
            tenants = tenants.filter(
                service_project_link__service__settings__domain=domain
            )
        return tenants

    def _validate_tenant_name(self, service_settings, tenant_name):
        neighbour_tenants = self._get_neighbour_tenants(service_settings)
        existing_tenant_names = [service_settings.options.get('tenant_name', 'admin')] +\
            list(neighbour_tenants.values_list('name', flat=True))
        if tenant_name in existing_tenant_names:
            raise serializers.ValidationError({
                'name': _('Name "%s" is already registered. Please choose another one.' % tenant_name),
            })

    def _validate_username(self, service_settings, username):
        neighbour_tenants = self._get_neighbour_tenants(service_settings)
        existing_usernames = [service_settings.username] + \
            list(neighbour_tenants.values_list('user_username', flat=True))
        if username in existing_usernames:
            raise serializers.ValidationError({
                'user_username': _('Name "%s" is already registered. Please choose another one.') % username
            })

        blacklisted_usernames = service_settings.options.get(
            'blacklisted_usernames', settings.WALDUR_OPENSTACK['DEFAULT_BLACKLISTED_USERNAMES'])
        if username in blacklisted_usernames:
            raise serializers.ValidationError({
                'user_username': _('Name "%s" cannot be used as tenant user username.') % username
            })

    def validate(self, attrs):
        attrs = super(TenantSerializer, self).validate(attrs)

        if not self.instance:
            self._validate_service_project_link(attrs['service_project_link'])

        self.validate_security_groups_configuration()

        if self.instance is not None:
            service_settings = self.instance.service_project_link.service.settings
        else:
            service_settings = attrs['service_project_link'].service.settings

        # validate tenant name
        if self.instance is not None and attrs.get('name'):
            if self.instance.name != attrs['name']:
                self._validate_tenant_name(service_settings, attrs['name'])
        else:
            self._validate_tenant_name(service_settings, attrs['name'])

        # username generation/validation
        if self.instance is not None or not settings.WALDUR_OPENSTACK['TENANT_CREDENTIALS_VISIBLE']:
            return attrs
        else:
            if not attrs.get('user_username'):
                attrs['user_username'] = models.Tenant.generate_username(attrs['name'])

            self._validate_username(service_settings, attrs.get('user_username'))

        return attrs

    def create(self, validated_data):
        spl = validated_data['service_project_link']
        # get availability zone from service settings if it is not defined
        if not validated_data.get('availability_zone'):
            validated_data['availability_zone'] = spl.service.settings.get_option('availability_zone') or ''
        # init tenant user username(if not defined) and password
        slugified_name = slugify(validated_data['name'])[:25]
        if not validated_data.get('user_username'):
            validated_data['user_username'] = models.Tenant.generate_username(validated_data['name'])
        validated_data['user_password'] = core_utils.pwgen()

        subnet_cidr = validated_data.pop('subnet_cidr')
        with transaction.atomic():
            tenant = super(TenantSerializer, self).create(validated_data)
            network = models.Network.objects.create(
                name=slugified_name + '-int-net',
                description=_('Internal network for tenant %s') % tenant.name,
                tenant=tenant,
                service_project_link=tenant.service_project_link,
            )
            models.SubNet.objects.create(
                name=slugified_name + '-sub-net',
                description=_('SubNet for tenant %s internal network') % tenant.name,
                network=network,
                service_project_link=tenant.service_project_link,
                cidr=subnet_cidr,
                allocation_pools=_generate_subnet_allocation_pool(subnet_cidr),
                dns_nameservers=spl.service.settings.options.get('dns_nameservers', [])
            )

            nc_settings = getattr(settings, 'WALDUR_OPENSTACK', {})
            config_groups = copy.deepcopy(nc_settings.get('DEFAULT_SECURITY_GROUPS', []))

            for group in config_groups:
                sg_name = group.get('name')
                sg_description = group.get('description', None)
                sg = models.SecurityGroup.objects.get_or_create(
                    service_project_link=tenant.service_project_link,
                    tenant=tenant,
                    description=sg_description,
                    name=sg_name)[0]

                for rule in group.get('rules'):
                    if 'icmp_type' in rule:
                        rule['from_port'] = rule.pop('icmp_type')
                    if 'icmp_code' in rule:
                        rule['to_port'] = rule.pop('icmp_code')

                    try:
                        rule = models.SecurityGroupRule(security_group=sg, **rule)
                        rule.full_clean()
                    except serializers.ValidationError as e:
                        logger.error('Failed to create rule for security group %s: %s.' % (sg_name, e))
                    else:
                        rule.save()

        return tenant


class _NestedSubNetSerializer(serializers.ModelSerializer):

    class Meta(object):
        model = models.SubNet
        fields = ('name', 'description', 'cidr', 'gateway_ip', 'allocation_pools', 'ip_version', 'enable_dhcp')


class NetworkSerializer(structure_serializers.BaseResourceActionSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstack-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        read_only=True)
    subnets = _NestedSubNetSerializer(many=True, read_only=True)
    tenant_name = serializers.CharField(source='tenant.name', read_only=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Network
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'tenant', 'tenant_name', 'is_external', 'type', 'segmentation_id', 'subnets')
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'tenant', 'is_external', 'type', 'segmentation_id', 'service_settings', 'project')
        extra_kwargs = dict(
            tenant={'lookup_field': 'uuid', 'view_name': 'openstack-tenant-detail'},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        attrs['tenant'] = tenant = self.context['view'].get_object()
        attrs['service_project_link'] = tenant.service_project_link
        return super(NetworkSerializer, self).validate(attrs)


class SubNetSerializer(structure_serializers.BaseResourceActionSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='openstack-detail',
        read_only=True,
        lookup_field='uuid')
    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='openstack-spl-detail',
        read_only=True)
    cidr = serializers.CharField(
        validators=[subnet_cidr_validator], required=False, initial='192.168.42.0/24', label='CIDR')
    allocation_pools = serializers.JSONField(read_only=True)
    network_name = serializers.CharField(source='network.name', read_only=True)
    tenant = serializers.HyperlinkedRelatedField(
        source='network.tenant',
        view_name='openstack-tenant-detail',
        read_only=True,
        lookup_field='uuid')
    tenant_name = serializers.CharField(source='network.tenant.name', read_only=True)
    dns_nameservers = serializers.JSONField(read_only=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.SubNet
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'tenant', 'tenant_name', 'network', 'network_name', 'cidr',
            'gateway_ip', 'disable_gateway', 'allocation_pools', 'ip_version', 'enable_dhcp', 'dns_nameservers')
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + ('cidr',)
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'tenant', 'network', 'ip_version', 'enable_dhcp', 'service_settings', 'project')
        extra_kwargs = dict(
            network={'lookup_field': 'uuid', 'view_name': 'openstack-network-detail'},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def validate(self, attrs):
        if attrs.get('disable_gateway') and attrs.get('gateway_ip'):
            raise serializers.ValidationError(
                _('These parameters are mutually exclusive: disable_gateway and gateway_ip.'))

        if self.instance is None:
            attrs['network'] = network = self.context['view'].get_object()
            if network.subnets.count() >= 1:
                raise serializers.ValidationError(_('Internal network cannot have more than one subnet.'))
            if 'cidr' not in attrs:
                attrs['cidr'] = '192.168.42.0/24'
            cidr = attrs['cidr']
            if models.SubNet.objects.filter(cidr=cidr, network__tenant=network.tenant).exists():
                raise serializers.ValidationError(_('Subnet with cidr "%s" is already registered') % cidr)

            attrs['service_project_link'] = network.service_project_link
            options = network.service_project_link.service.settings.options
            attrs['allocation_pools'] = _generate_subnet_allocation_pool(cidr)
            attrs['dns_nameservers'] = options.get('dns_nameservers', [])

        return attrs


def _generate_subnet_allocation_pool(cidr):
    first_octet, second_octet, third_octet, _ = cidr.split('.', 3)
    subnet_settings = settings.WALDUR_OPENSTACK['SUBNET']
    format_data = {'first_octet': first_octet, 'second_octet': second_octet, 'third_octet': third_octet}
    return [{
        'start': subnet_settings['ALLOCATION_POOL_START'].format(**format_data),
        'end': subnet_settings['ALLOCATION_POOL_END'].format(**format_data),
    }]


class TenantChangePasswordSerializer(serializers.Serializer):
    user_password = serializers.CharField(max_length=50,
                                          allow_blank=True,
                                          validators=[password_validation.validate_password],
                                          help_text=_('New tenant user password.'))

    def validate_user_password(self, user_password):
        if self.instance.user_password == user_password:
            raise serializers.ValidationError(_('New password cannot match the old password.'))

        return user_password

    def update(self, tenant, validated_data):
        tenant.user_password = validated_data['user_password']
        tenant.save(update_fields=['user_password'])
        return tenant



import base64
import datetime
from calendar import monthrange
from decimal import Decimal

import pdfkit
from django.conf import settings
from django.template.loader import render_to_string
from django.utils import timezone

from waldur_core.core import utils as core_utils
from waldur_mastermind.common.mixins import UnitPriceMixin


def get_current_month():
    return timezone.now().month


def get_current_year():
    return timezone.now().year


def get_current_month_end():
    return core_utils.month_end(timezone.now())


def get_current_month_start():
    return core_utils.month_start(timezone.now())


def get_full_days(start, end):
    seconds_in_day = 24 * 60 * 60
    full_days, extra_seconds = divmod((end - start).total_seconds(), seconds_in_day)
    if extra_seconds > 0:
        full_days += 1

    return int(full_days)


def get_current_month_days():
    now = timezone.now()
    range = monthrange(now.year, now.month)
    return range[1]


def get_full_hours(start, end):
    seconds_in_hour = 60 * 60
    full_hours, extra_seconds = divmod((end - start).total_seconds(), seconds_in_hour)
    if extra_seconds > 0:
        full_hours += 1

    return int(full_hours)


def check_past_date(year, month, day=None):
    day = day or 1

    try:
        return datetime.date(year=int(year), month=int(month), day=int(day)) <= timezone.now().date()
    except ValueError:
        return False


def parse_period(attrs, use_default=True):
    year = use_default and get_current_year() or None
    month = use_default and get_current_month() or None

    try:
        year = int(attrs.get('year', ''))
        month = int(attrs.get('month', ''))
    except ValueError:
        pass

    return year, month


def get_previous_month():
    date = timezone.now()
    month, year = (date.month - 1, date.year) if date.month != 1 else (12, date.year - 1)
    return datetime.date(year, month, 1)


def filter_invoice_items(items):
    return [item for item in items if item.total > 0]


def create_invoice_pdf(invoice):
    all_items = filter_invoice_items(invoice.items)
    logo_path = settings.WALDUR_CORE['SITE_LOGO']
    if logo_path:
        with open(logo_path, 'r') as image_file:
            deployment_logo = base64.b64encode(image_file.read())
    else:
        deployment_logo = None

    context = dict(
        invoice=invoice,
        issuer_details=settings.WALDUR_INVOICES['ISSUER_DETAILS'],
        currency=settings.WALDUR_CORE['CURRENCY_NAME'],
        deployment_logo=deployment_logo,
        items=all_items,
    )
    html = render_to_string('invoices/invoice.html', context)
    pdf = pdfkit.from_string(html, False)
    invoice.file = base64.b64encode(pdf)
    invoice.save()


def get_price_per_day(price, unit):
    if unit == UnitPriceMixin.Units.PER_DAY:
        return price
    elif unit == UnitPriceMixin.Units.PER_MONTH:
        return price / Decimal(30)
    elif unit == UnitPriceMixin.Units.PER_HALF_MONTH:
        return price / Decimal(15)
    elif unit == UnitPriceMixin.Units.PER_HOUR:
        return price * 24
    else:
        return price



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-29 14:33
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import re
import taggit.managers
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.logging.loggers


class Migration(migrations.Migration):

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('waldur_azure', '0003_redesign'),
    ]

    operations = [
        migrations.CreateModel(
            name='StorageAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('name', models.CharField(max_length=24, validators=[django.core.validators.RegexValidator(message='The name can contain only letters and numbers.', regex=re.compile(b'[a-z][a-z0-9]{2,23}$'))])),
                ('resource_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.ResourceGroup')),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_azure.AzureServiceProjectLink')),
                ('tags', taggit.managers.TaggableManager(blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.core.models.BackendModelMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-27 10:00
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0057_category_backend_id'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='terms_of_service',
            field=models.CharField(blank=True, default='', max_length=500),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-09-10 11:05
from __future__ import unicode_literals

from decimal import Decimal
from django.conf import settings
import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.structure.models


class Migration(migrations.Migration):
    replaces = [(b'support', '0001_initial'), (b'support', '0002_comment_and_support_user'),
                (b'support', '0003_additional_issue_fields'), (b'support', '0004_issue_users'),
                (b'support', '0005_issue_first_response_sla'), (b'support', '0006_offering'),
                (b'support', '0007_offering_update'), (b'support', '0008_add_product_code'),
                (b'support', '0009_offering_article_code'), (b'support', '0010_add_unit_price'),
                (b'support', '0011_remove_price'), (b'support', '0012_attachment'),
                (b'support', '0013_add_more_backend_fields'), (b'support', '0014_unique_backend_id_for_attachment'),
                (b'support', '0015_offering_report'), (b'support', '0016_add_states'),
                (b'support', '0017_unique_backend_id'), (b'support', '0018_offering_terminated_at'),
                (b'support', '0019_template_templateattachment'), (b'support', '0020_protect_offering_issue'),
                (b'support', '0021_protect_issue_user'), (b'support', '0022_offering_template'),
                (b'support', '0023_offering_template_migrate_data'), (b'support', '0024_offeringplan'),
                (b'support', '0025_offering_plan_data'), (b'support', '0026_remove_type_protect_template'),
                (b'support', '0027_non_unique_template_name'),
                (b'support', '0028_ignoredissuestatus_templatestatusnotification'),
                (b'support', '0029_issue_resolution_date'), (b'support', '0030_issue_template'),
                (b'support', '0031_supportcustomer'), (b'support', '0032_requesttype'),
                (b'support', '0033_issuestatus'), (b'support', '0034_bootstrap_statuses'),
                (b'support', '0035_offering_plan'), (b'support', '0036_offering_ordering'),
                (b'support', '0037_priority')]

    initial = True

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('file', models.FileField(upload_to='support_attachments')),
                ('backend_id', models.CharField(max_length=255, unique=True)),
                ('mime_type', models.CharField(blank=True, max_length=100, verbose_name='MIME type')),
                ('file_size', models.PositiveIntegerField(blank=True, null=True, verbose_name='Filesize, B')),
                ('thumbnail', models.FileField(blank=True, null=True, upload_to='support_attachments_thumbnails')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('description', models.TextField()),
                ('is_public', models.BooleanField(default=True)),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True)),
            ],
            options={
                'ordering': ['-created'],
            },
            bases=(waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='IgnoredIssueStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, unique=True, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
            ],
        ),
        migrations.CreateModel(
            name='Issue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('key', models.CharField(blank=True, max_length=255)),
                ('type', models.CharField(max_length=255)),
                ('link', models.URLField(blank=True, help_text='Link to issue in support system.', max_length=255)),
                ('summary', models.CharField(max_length=255)),
                ('description', models.TextField(blank=True)),
                ('deadline', models.DateTimeField(blank=True, null=True)),
                ('impact', models.CharField(blank=True, max_length=255)),
                ('status', models.CharField(max_length=255)),
                ('resolution', models.CharField(blank=True, max_length=255)),
                ('priority', models.CharField(blank=True, max_length=255)),
                ('resource_object_id', models.PositiveIntegerField(null=True)),
                ('first_response_sla', models.DateTimeField(blank=True, null=True)),
                ('resolution_date', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'ordering': ['-created'],
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, waldur_core.core.models.BackendModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='IssueStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Status name in Jira.', max_length=255, unique=True)),
                ('type', django_fsm.FSMIntegerField(choices=[(0, 'Resolved'), (1, 'Canceled')], default=0)),
            ],
            options={
                'verbose_name': 'Issue status',
                'verbose_name_plural': 'Issue statuses',
            },
        ),
        migrations.CreateModel(
            name='Offering',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('unit_price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('unit', models.CharField(choices=[('month', 'Per month'), ('half_month', 'Per half month'), ('day', 'Per day'), ('hour', 'Per hour'), ('quantity', 'Quantity')], default='day', max_length=30)),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
                ('state', models.CharField(choices=[('requested', 'Requested'), ('ok', 'OK'), ('terminated', 'Terminated')], default='requested', max_length=30)),
                ('report', waldur_core.core.fields.JSONField(blank=True)),
                ('terminated_at', models.DateTimeField(blank=True, editable=False, null=True)),
                ('issue', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='support.Issue')),
            ],
            options={
                'ordering': ['-created'],
                'verbose_name': 'Request',
                'verbose_name_plural': 'Requests',
            },
            bases=(waldur_core.structure.models.StructureLoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OfferingPlan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('unit_price', models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
                ('unit', models.CharField(choices=[('month', 'Per month'), ('half_month', 'Per half month'), ('day', 'Per day'), ('hour', 'Per hour'), ('quantity', 'Quantity')], default='day', max_length=30)),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='OfferingTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('name', models.CharField(max_length=150, verbose_name='name')),
                ('config', django.contrib.postgres.fields.jsonb.JSONField()),
                ('sort_order', models.PositiveSmallIntegerField(default=1)),
            ],
            options={
                'ordering': ['sort_order', 'name'],
            },
        ),
        migrations.CreateModel(
            name='Priority',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('icon_url', models.URLField(blank=True, verbose_name='icon url')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(blank=True, max_length=255)),
            ],
            options={
                'verbose_name': 'Priority',
                'verbose_name_plural': 'Priorities',
            },
        ),
        migrations.CreateModel(
            name='RequestType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.IntegerField(unique=True)),
                ('issue_type_name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SupportCustomer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('backend_id', models.CharField(max_length=255, unique=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='SupportUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(blank=True, max_length=255, null=True, unique=True)),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('native_name', models.CharField(blank=True, max_length=150)),
                ('description', models.TextField()),
                ('native_description', models.TextField(blank=True)),
                ('issue_type', models.CharField(choices=[('INFORMATIONAL', 'Informational'), ('SERVICE_REQUEST', 'Service request'), ('CHANGE_REQUEST', 'Change request'), ('INCIDENT', 'Incident')], default='INFORMATIONAL', max_length=30)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TemplateAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('file', models.FileField(upload_to='support_template_attachments')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='support.Template')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TemplateStatusNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(max_length=255, unique=True, validators=[waldur_core.core.validators.validate_name])),
                ('html', models.TextField(validators=[waldur_core.core.validators.validate_name])),
                ('text', models.TextField(validators=[waldur_core.core.validators.validate_name])),
                ('subject', models.CharField(max_length=255, validators=[waldur_core.core.validators.validate_name])),
            ],
        ),
        migrations.AddField(
            model_name='offeringplan',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='plans', to='support.OfferingTemplate'),
        ),
        migrations.AddField(
            model_name='offering',
            name='plan',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='support.OfferingPlan'),
        ),
        migrations.AddField(
            model_name='offering',
            name='project',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='structure.Project'),
        ),
        migrations.AddField(
            model_name='offering',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='support.OfferingTemplate'),
        ),
        migrations.AddField(
            model_name='issue',
            name='assignee',
            field=models.ForeignKey(blank=True, help_text='Help desk user who will implement the issue', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='issues', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='issue',
            name='caller',
            field=models.ForeignKey(help_text='Waldur user who has reported the issue.', on_delete=django.db.models.deletion.PROTECT, related_name='created_issues', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='issue',
            name='customer',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='issues', to='structure.Customer', verbose_name='organization'),
        ),
        migrations.AddField(
            model_name='issue',
            name='project',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='issues', to='structure.Project'),
        ),
        migrations.AddField(
            model_name='issue',
            name='reporter',
            field=models.ForeignKey(blank=True, help_text='Help desk user who have created the issue that is reported by caller.', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='reported_issues', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='issue',
            name='resource_content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='issue',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='issues', to='support.Template'),
        ),
        migrations.AddField(
            model_name='comment',
            name='author',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='comment',
            name='issue',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments', to='support.Issue'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='attachment',
            name='issue',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='support.Issue'),
        ),
        migrations.AlterUniqueTogether(
            name='comment',
            unique_together=set([('backend_id', 'issue')]),
        ),
    ]
